"""Initializes the .fmu directory."""

from pathlib import Path
from typing import Any, Final

from fmu.datamodels.fmu_results.global_configuration import GlobalConfiguration

from ._fmu_dir import ProjectFMUDirectory, UserFMUDirectory
from ._logging import null_logger
from ._readme_texts import PROJECT_README_CONTENT, USER_README_CONTENT
from ._resources.lock_manager import DEFAULT_LOCK_TIMEOUT
from .models.project_config import ProjectConfig

logger: Final = null_logger(__name__)


def _create_fmu_directory(base_path: Path) -> None:
    """Creates the .fmu directory.

    Args:
        base_path: Base directory where .fmu should be created

    Raises:
        FileNotFoundError: If base_path doesn't exist
        FileExistsError: If .fmu exists
    """
    logger.debug(f"Creating .fmu directory in '{base_path}'")

    if not base_path.exists():
        raise FileNotFoundError(
            f"Base path '{base_path}' does not exist. Expected the root "
            "directory of an FMU project."
        )

    fmu_dir = base_path / ".fmu"
    if fmu_dir.exists():
        if fmu_dir.is_dir():
            raise FileExistsError(f"{fmu_dir} already exists")
        raise FileExistsError(f"{fmu_dir} exists but is not a directory")

    fmu_dir.mkdir()
    logger.debug(f"Created .fmu directory at '{fmu_dir}'")


def init_fmu_directory(
    base_path: str | Path,
    config_data: ProjectConfig | dict[str, Any] | None = None,
    global_config: GlobalConfiguration | None = None,
    *,
    lock_timeout_seconds: int = DEFAULT_LOCK_TIMEOUT,
) -> ProjectFMUDirectory:
    """Creates and initializes a .fmu directory.

    Also initializes a configuration file if configuration data is provided through the
    function.

    Args:
        base_path: Directory where .fmu should be created.
        config_data: Optional ProjectConfig instance or dictionary with configuration
          data.
        global_config: Optional GlobaConfiguration instance with existing global config
          data.
        lock_timeout_seconds: Lock expiration time in seconds. Default 20 minutes.

    Returns:
        Instance of FMUDirectory

    Raises:
        FileExistsError: If .fmu exists
        FileNotFoundError: If base_path doesn't exist
        PermissionError: If the user lacks permission to create directories
        ValidationError: If config_data fails validationg
    """
    logger.debug("Initializing .fmu directory")
    base_path = Path(base_path)

    _create_fmu_directory(base_path)

    fmu_dir = ProjectFMUDirectory(
        base_path,
        lock_timeout_seconds=lock_timeout_seconds,
    )
    fmu_dir.write_text_file("README", PROJECT_README_CONTENT)

    fmu_dir.config.reset()
    if config_data:
        if isinstance(config_data, ProjectConfig):
            config_data = config_data.model_dump()
        fmu_dir.update_config(config_data)

    if global_config:
        for key, value in global_config.model_dump().items():
            fmu_dir.set_config_value(key, value)

    logger.info(f"Successfully initialized .fmu directory at '{fmu_dir}'")
    return fmu_dir


def init_user_fmu_directory(
    *,
    lock_timeout_seconds: int = DEFAULT_LOCK_TIMEOUT,
) -> UserFMUDirectory:
    """Creates and initializes a user's $HOME/.fmu directory.

    Args:
        lock_timeout_seconds: Lock expiration time in seconds. Default 20 minutes.

    Returns:
        Instance of FMUDirectory

    Raises:
        FileExistsError: If .fmu exists
        FileNotFoundError: If base_path doesn't exist
        PermissionError: If the user lacks permission to create directories
        ValidationError: If config_data fails validationg
    """
    logger.debug("Initializing .fmu directory")

    _create_fmu_directory(Path.home())

    fmu_dir = UserFMUDirectory(lock_timeout_seconds=lock_timeout_seconds)
    fmu_dir.write_text_file("README", USER_README_CONTENT)

    fmu_dir.config.reset()
    logger.debug(f"Successfully initialized .fmu directory at '{fmu_dir}'")
    return fmu_dir



"""The generic configuration file in a .fmu directory."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Final, Self

from fmu.settings._logging import null_logger
from fmu.settings.models.project_config import ProjectConfig
from fmu.settings.models.user_config import UserConfig

from .pydantic_resource_manager import (
    MutablePydanticResourceManager,
)

if TYPE_CHECKING:
    # Avoid circular dependency for type hint in __init__ only
    from fmu.settings._fmu_dir import (
        ProjectFMUDirectory,
        UserFMUDirectory,
    )

logger: Final = null_logger(__name__)


class ProjectConfigManager(MutablePydanticResourceManager[ProjectConfig]):
    """Manages the .fmu configuration file in a project."""

    def __init__(self: Self, fmu_dir: ProjectFMUDirectory) -> None:
        """Initializes the ProjectConfig resource manager."""
        super().__init__(fmu_dir, ProjectConfig)

    @property
    def relative_path(self: Self) -> Path:
        """Returns the relative path to the config file."""
        return Path("config.json")


class UserConfigManager(MutablePydanticResourceManager[UserConfig]):
    """Manages the .fmu configuration file in a user's home directory."""

    def __init__(self: Self, fmu_dir: UserFMUDirectory) -> None:
        """Initializes the UserConfig resource manager."""
        super().__init__(fmu_dir, UserConfig)

    @property
    def relative_path(self: Self) -> Path:
        """Returns the relative path to the config file."""
        return Path("config.json")



from __future__ import annotations

import os
import socket
from datetime import UTC, datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Self

from pydantic import BaseModel

from fmu.settings._resources.log_manager import LogManager
from fmu.settings.models._enums import ChangeType
from fmu.settings.models.change_info import ChangeInfo
from fmu.settings.models.log import Log, LogFileName

if TYPE_CHECKING:
    # Avoid circular dependency for type hint in __init__ only
    from fmu.settings._fmu_dir import (
        FMUDirectoryBase,
    )


class ChangelogManager(LogManager[ChangeInfo]):
    """Manages the .fmu changelog file."""

    def __init__(self: Self, fmu_dir: FMUDirectoryBase) -> None:
        """Initializes the Change log resource manager."""
        super().__init__(fmu_dir, Log[ChangeInfo])

    @property
    def relative_path(self: Self) -> Path:
        """Returns the relative path to the log file."""
        return Path("logs") / LogFileName.changelog

    def log_update_to_changelog(
        self: Self,
        updates: dict[str, Any],
        old_resource_dict: dict[str, Any],
        relative_path: Path,
    ) -> None:
        """Logs the update of a resource to the changelog."""
        _MISSING_KEY = object()
        for key, new_value in updates.items():
            change_type = ChangeType.update
            if "." in key:
                old_value = self._get_dot_notation_key(
                    resource_dict=old_resource_dict, key=key, default=_MISSING_KEY
                )
            else:
                old_value = old_resource_dict.get(key, _MISSING_KEY)

            if old_value != _MISSING_KEY:
                old_value_string = (
                    str(old_value.model_dump())
                    if isinstance(old_value, BaseModel)
                    else str(old_value)
                )
                new_value_string = (
                    str(new_value.model_dump())
                    if isinstance(new_value, BaseModel)
                    else str(new_value)
                )
                change_string = (
                    f"Updated field '{key}'. Old value: {old_value_string}"
                    f" -> New value: {new_value_string}"
                )
            else:
                change_type = ChangeType.add
                new_value_string = (
                    str(new_value.model_dump())
                    if isinstance(new_value, BaseModel)
                    else str(new_value)
                )
                change_string = f"Added field '{key}'. New value: {new_value_string}"

            change_entry = ChangeInfo(
                timestamp=datetime.now(UTC),
                change_type=change_type,
                user=os.getenv("USER", "unknown"),
                path=self.fmu_dir.path,
                change=change_string,
                hostname=socket.gethostname(),
                file=str(relative_path),
                key=key,
            )
            self.add_log_entry(change_entry)



"""Root configuration for pytest."""

from collections.abc import Callable
from datetime import UTC, datetime
from pathlib import Path
from typing import Any
from unittest.mock import patch
from uuid import uuid4

import pytest
import yaml
from fmu.datamodels.fmu_results import fields
from fmu.datamodels.fmu_results.enums import Classification
from fmu.datamodels.fmu_results.global_configuration import (
    Access,
    GlobalConfiguration,
    Stratigraphy,
    StratigraphyElement,
)
from pytest import MonkeyPatch

from fmu.settings._fmu_dir import ProjectFMUDirectory, UserFMUDirectory
from fmu.settings._init import init_fmu_directory, init_user_fmu_directory
from fmu.settings._version import __version__
from fmu.settings.models.project_config import ProjectConfig
from fmu.settings.models.user_config import UserConfig


@pytest.fixture
def unix_epoch_utc() -> datetime:
    """Returns a fixed datetime used in testing."""
    return datetime(1970, 1, 1, 0, 0, tzinfo=UTC)


@pytest.fixture
def config_dict(unix_epoch_utc: datetime) -> dict[str, Any]:
    """A dictionary representing a .fmu config."""
    return {
        "version": __version__,
        "created_at": unix_epoch_utc,
        "created_by": "user",
        "cache_max_revisions": 5,
        "masterdata": None,
        "model": None,
        "access": None,
    }


@pytest.fixture
def masterdata_dict() -> dict[str, Any]:
    """Example masterdata from SMDA."""
    return {
        "smda": {
            "country": [
                {
                    "identifier": "Norway",
                    "uuid": "ad214d85-8a1d-19da-e053-c918a4889309",
                }
            ],
            "discovery": [
                {
                    "short_identifier": "DROGON",
                    "uuid": "ad214d85-8a1d-19da-e053-c918a4889309",
                }
            ],
            "field": [
                {
                    "identifier": "DROGON",
                    "uuid": "ad214d85-8a1d-19da-e053-c918a4889309",
                }
            ],
            "coordinate_system": {
                "identifier": "ST_WGS84_UTM37N_P32637",
                "uuid": "ad214d85-dac7-19da-e053-c918a4889309",
            },
            "stratigraphic_column": {
                "identifier": "DROGON_HAS_NO_STRATCOLUMN",
                "uuid": "ad214d85-8a1d-19da-e053-c918a4889309",
            },
        }
    }


@pytest.fixture
def model_dict() -> dict[str, Any]:
    """Example model information."""
    return {
        "name": "Drogon",
        "revision": "21.0.0",
        "description": None,
    }


@pytest.fixture
def access_dict() -> dict[str, Any]:
    """Example access information."""
    return {
        "asset": {"name": "Drogon"},
        "classification": "internal",
    }


@pytest.fixture
def stratigraphy_dict() -> dict[str, Any]:
    """Example stratigraphy information."""
    return {
        "MSL": {
            "stratigraphic": False,
            "name": "MSL",
        },
        "Seabase": {
            "stratigraphic": False,
            "name": "Seabase",
        },
        "TopVolantis": {
            "stratigraphic": True,
            "name": "VOLANTIS GP. Top",
            "alias": ["TopVOLANTIS", "TOP_VOLANTIS"],
            "stratigraphic_alias": ["TopValysar", "Valysar Fm. Top"],
        },
        "TopTherys": {"stratigraphic": True, "name": "Therys Fm. Top"},
        "TopVolon": {"stratigraphic": True, "name": "Volon Fm. Top"},
        "BaseVolon": {"stratigraphic": True, "name": "Volon Fm. Base"},
        "BaseVolantis": {"stratigraphic": True, "name": "VOLANTIS GP. Base"},
        "Mantle": {"stratigraphic": False, "name": "Mantle"},
        "Above": {"stratigraphic": False, "name": "Above"},
        "Valysar": {"stratigraphic": True, "name": "Valysar Fm."},
        "Therys": {"stratigraphic": True, "name": "Therys Fm."},
        "Volon": {"stratigraphic": True, "name": "Volon Fm."},
        "Below": {"stratigraphic": False, "name": "Below"},
    }


@pytest.fixture
def global_variables_without_masterdata() -> dict[str, Any]:
    """Example global_variables.yml file without masterdata."""
    return {
        "global": {
            "dates": ["2018-01-01", "2018-07-01", "2019-07-01", "2020-07-01"],
        },
    }


@pytest.fixture
def global_variables_with_masterdata(
    masterdata_dict: dict[str, Any],
    access_dict: dict[str, Any],
    model_dict: dict[str, Any],
    stratigraphy_dict: dict[str, Any],
    global_variables_without_masterdata: dict[str, Any],
) -> dict[str, Any]:
    """Example global_variables.yml file with masterdata."""
    return {
        "masterdata": masterdata_dict,
        "access": access_dict,
        "model": model_dict,
        "stratigraphy": stratigraphy_dict,
        **global_variables_without_masterdata,
    }


@pytest.fixture
def fmuconfig_with_input(  # noqa: PLR0913 too many args
    tmp_path: Path,
    monkeypatch: MonkeyPatch,
    masterdata_dict: dict[str, Any],
    access_dict: dict[str, Any],
    model_dict: dict[str, Any],
    stratigraphy_dict: dict[str, Any],
    global_variables_without_masterdata: dict[str, Any],
) -> Path:
    """Creates an fmuconfig/ path and input directory.

    Returns:
        tmp_path.
    """
    fmuconfig_input = tmp_path / "fmuconfig/input"
    fmuconfig_input.mkdir(parents=True, exist_ok=True)

    with open(fmuconfig_input / "_masterdata.yml", "w") as f:
        yaml.safe_dump(masterdata_dict, f)
    with open(fmuconfig_input / "_access.yml", "w") as f:
        yaml.safe_dump(access_dict, f)
    with open(fmuconfig_input / "_stratigraphy.yml", "w") as f:
        yaml.safe_dump(stratigraphy_dict, f)

    global_config_dict = {
        **global_variables_without_masterdata,
        "model": model_dict,
        "masterdata": "!include _masterdata.yml",
        "access": "!include _access.yml",
        "stratigraphy": "!include _stratigraphy.yml",
    }
    # Remove quotes around strings. PyYAML quotes string beginning with ! by default as
    # they indicate a yaml tag, for which fmu-config has defined one in this case.
    global_config_yml = yaml.dump(
        global_config_dict, default_style=None, default_flow_style=False
    ).replace("'", "")

    with open(fmuconfig_input / "global_master_config.yml", "w") as f:
        f.write(global_config_yml)

    return tmp_path


@pytest.fixture
def fmuconfig_with_output(  # noqa: PLR0913 too many args
    fmuconfig_with_input: Path,
    global_variables_with_masterdata: dict[str, Any],
) -> Path:
    """Creates an fmuconfig/ path and input/output directory.

    Returns:
        tmp_path
    """
    fmuconfig_output = fmuconfig_with_input / "fmuconfig/output"
    fmuconfig_output.mkdir(parents=True, exist_ok=True)

    with open(fmuconfig_output / "global_variables.yml", "w") as f:
        yaml.safe_dump(
            global_variables_with_masterdata,
            f,
            default_style=None,
            default_flow_style=False,
        )

    return fmuconfig_with_input


@pytest.fixture
def generate_strict_valid_globalconfiguration() -> Callable[[], GlobalConfiguration]:
    """Generates a global configuration that is valid, but can switch particular models.

    All values are left empty by default.
    """

    def _generate_cfg(  # noqa: PLR0913
        *,
        classification: Classification | None = Classification.internal,
        asset: fields.Asset | None = None,
        coordinate_system: fields.CoordinateSystem | None = None,
        stratigraphic_column: fields.StratigraphicColumn | None = None,
        country_items: list[fields.CountryItem] | None = None,
        discovery_items: list[fields.DiscoveryItem] | None = None,
        field_items: list[fields.FieldItem] | None = None,
        model: fields.Model | None = None,
    ) -> GlobalConfiguration:
        return GlobalConfiguration(
            access=Access(
                asset=asset or fields.Asset(name=""), classification=classification
            ),
            masterdata=fields.Masterdata(
                smda=fields.Smda(
                    coordinate_system=(
                        coordinate_system
                        or fields.CoordinateSystem(identifier="", uuid=uuid4())
                    ),
                    stratigraphic_column=(
                        stratigraphic_column
                        or fields.StratigraphicColumn(identifier="", uuid=uuid4())
                    ),
                    country=country_items or [],
                    discovery=discovery_items or [],
                    field=field_items or [],
                )
            ),
            model=model or fields.Model(name="", revision=""),
            stratigraphy=Stratigraphy(
                {"MSL": StratigraphyElement(name="MSL", stratigraphic=False)}
            ),
        )

    return _generate_cfg


@pytest.fixture
def config_dict_with_masterdata(
    unix_epoch_utc: datetime,
    masterdata_dict: dict[str, Any],
    model_dict: dict[str, Any],
) -> dict[str, Any]:
    """A dictionary representing a .fmu config."""
    return {
        "version": __version__,
        "created_at": unix_epoch_utc,
        "created_by": "user",
        "cache_max_revisions": 5,
        "masterdata": masterdata_dict,
        "model": model_dict,
    }


@pytest.fixture
def config_model(config_dict: dict[str, Any]) -> ProjectConfig:
    """A ProjectConfig model representing a .fmu config file."""
    return ProjectConfig.model_validate(config_dict)


@pytest.fixture
def config_model_with_masterdata(
    config_dict_with_masterdata: dict[str, Any],
) -> ProjectConfig:
    """A ProjectConfig model representing a .fmu config file."""
    return ProjectConfig.model_validate(config_dict_with_masterdata)


@pytest.fixture
def user_config_dict(unix_epoch_utc: datetime) -> dict[str, Any]:
    """A dictionary representing a .fmu user config."""
    return {
        "version": __version__,
        "created_at": unix_epoch_utc,
        "cache_max_revisions": 5,
        "user_api_keys": {
            "smda_subscription": None,
        },
        "recent_project_directories": [],
    }


@pytest.fixture
def user_config_model(user_config_dict: dict[str, Any]) -> UserConfig:
    """A UserConfig model representing a .fmu config file."""
    return UserConfig.model_validate(user_config_dict)


@pytest.fixture(scope="function")
def fmu_dir(tmp_path: Path, unix_epoch_utc: datetime) -> ProjectFMUDirectory:
    """Create an ProjectFMUDirectory instance for testing."""
    with (
        patch(
            "fmu.settings.models.project_config.getpass.getuser",
            return_value="user",
        ),
        patch("fmu.settings.models.project_config.datetime") as mock_datetime,
    ):
        mock_datetime.now.return_value = unix_epoch_utc
        mock_datetime.datetime.now.return_value = unix_epoch_utc
        return init_fmu_directory(tmp_path)


@pytest.fixture
def user_fmu_dir(tmp_path: Path, unix_epoch_utc: datetime) -> UserFMUDirectory:
    """Create an ProjectFMUDirectory instance for testing."""
    with (
        patch("pathlib.Path.home", return_value=tmp_path),
        patch("fmu.settings.models.user_config.datetime") as mock_datetime,
    ):
        mock_datetime.now.return_value = unix_epoch_utc
        mock_datetime.datetime.now.return_value = unix_epoch_utc

        return init_user_fmu_directory()



"""Tests for fmu.settings._create."""

import json
from datetime import UTC, datetime, timedelta
from pathlib import Path
from typing import Any
from unittest.mock import patch

import pytest

from fmu.settings import __version__
from fmu.settings._global_config import find_global_config
from fmu.settings._init import (
    _create_fmu_directory,
    init_fmu_directory,
    init_user_fmu_directory,
)
from fmu.settings._readme_texts import PROJECT_README_CONTENT, USER_README_CONTENT
from fmu.settings.models.project_config import ProjectConfig
from fmu.settings.models.user_config import UserConfig


def test_create_fmu_directory(tmp_path: Path) -> None:
    """Tests creating the .fmu directory raises appropriate exceptions."""
    with pytest.raises(FileNotFoundError, match="Base path '/foo' does not exist."):
        _create_fmu_directory(Path("/foo"))

    fmu_dir = tmp_path / ".fmu"
    fmu_dir.mkdir()
    with pytest.raises(FileExistsError, match=f"{fmu_dir} already exists"):
        _create_fmu_directory(tmp_path)

    fmu_dir.rmdir()
    fmu_dir.touch()
    with pytest.raises(
        FileExistsError, match=f"{fmu_dir} exists but is not a directory"
    ):
        _create_fmu_directory(tmp_path)


def test_init_fmu_directory_with_no_config_data(
    tmp_path: Path,
    unix_epoch_utc: datetime,
) -> None:
    """Tests initializing a .fmu directory with default settings."""
    with (
        patch(
            "fmu.settings.models.project_config.getpass.getuser",
            return_value="user",
        ),
    ):
        fmu_dir = init_fmu_directory(tmp_path)

    assert fmu_dir.path.exists()
    assert fmu_dir.path.is_dir()
    assert fmu_dir.path == tmp_path / ".fmu"

    config_file = fmu_dir.path / "config.json"
    assert config_file.exists()

    with open(config_file, encoding="utf-8") as f:
        config_json = json.load(f)

    assert config_json["version"] == __version__
    assert config_json["created_by"] == "user"
    assert config_json["created_at"] != str(unix_epoch_utc)

    created_at = datetime.fromisoformat(config_json["created_at"])
    now = datetime.now(UTC)
    one_min_ago = now - timedelta(minutes=1)
    assert one_min_ago <= created_at <= now


def test_create_fmu_directory_with_config_model(
    tmp_path: Path,
    config_model: ProjectConfig,
) -> None:
    """Tests initializing a .fmu directory with a ProjectConfig model."""
    config_model.version = "200.0.0"
    config_model.model_rebuild()
    fmu_dir = init_fmu_directory(tmp_path, config_model)
    config = fmu_dir.config.load()
    assert config.version == "200.0.0"


def test_create_fmu_directory_with_config_dict(
    tmp_path: Path,
    config_dict: dict[str, Any],
) -> None:
    """Tests initializing a .fmu directory with a ProjectConfig model."""
    config_dict["version"] = "200.0.0"
    fmu_dir = init_fmu_directory(tmp_path, config_dict)
    config = fmu_dir.config.load()
    assert config.version == "200.0.0"


def test_write_fmu_config_roundtrip(tmp_path: Path) -> None:
    """Tests that the FMU config writes correctly."""
    fmu_dir = init_fmu_directory(tmp_path)
    assert str(fmu_dir.config.path).endswith("config.json")
    with open(fmu_dir.config.path, encoding="utf-8") as f:
        config_data = json.loads(f.read())
    # Fails if invalid
    ProjectConfig.model_validate(config_data)


def test_write_fmu_config_with_global_config(fmuconfig_with_output: Path) -> None:
    """Tests creating an .fmu config with a global_config."""
    tmp_path = fmuconfig_with_output
    cfg = find_global_config(tmp_path, strict=False)
    assert cfg is not None
    fmu_dir = init_fmu_directory(tmp_path, global_config=cfg)
    config = fmu_dir.config.load()
    assert config.masterdata is not None
    assert config.masterdata.smda.field[0].identifier == "DROGON"

    assert config.model is not None
    assert config.model.name == "Drogon"

    assert config.access is not None
    assert config.access.asset.name == "Drogon"


def test_readme_is_written(tmp_path: Path, config_model: ProjectConfig) -> None:
    """Tests that the README is written when .fmu is initialized."""
    fmu_dir = init_fmu_directory(tmp_path, config_model)

    readme = fmu_dir.path / "README"
    assert readme.exists()
    assert readme.read_text() == PROJECT_README_CONTENT


def test_init_user_fmu_directory(
    tmp_path: Path,
    unix_epoch_utc: datetime,
) -> None:
    """Tests initializing a user .fmu directory with default settings."""
    with patch("pathlib.Path.home", return_value=tmp_path):
        fmu_dir = init_user_fmu_directory()

    assert fmu_dir.path.exists()
    assert fmu_dir.path.is_dir()
    assert fmu_dir.path == tmp_path / ".fmu"

    config_file = fmu_dir.path / "config.json"
    assert config_file.exists()

    with open(config_file, encoding="utf-8") as f:
        config_json = json.load(f)

    assert config_json["version"] == __version__
    assert config_json["created_at"] != str(unix_epoch_utc)
    assert config_json["user_api_keys"] == {"smda_subscription": None}
    assert config_json["recent_project_directories"] == []

    created_at = datetime.fromisoformat(config_json["created_at"])
    now = datetime.now(UTC)
    one_min_ago = now - timedelta(minutes=1)
    assert one_min_ago <= created_at <= now


def test_write_user_fmu_config_roundtrip(tmp_path: Path) -> None:
    """Tests that the FMU config writes correctly."""
    with patch("pathlib.Path.home", return_value=tmp_path):
        fmu_dir = init_user_fmu_directory()
    assert str(fmu_dir.config.path).endswith("config.json")
    with open(fmu_dir.config.path, encoding="utf-8") as f:
        config_data = json.loads(f.read())
    # Fails if invalid
    UserConfig.model_validate(config_data)


def test_user_readme_is_written(tmp_path: Path, config_model: ProjectConfig) -> None:
    """Tests that the README is written when .fmu is initialized."""
    with patch("pathlib.Path.home", return_value=tmp_path):
        fmu_dir = init_user_fmu_directory()

    readme = fmu_dir.path / "README"
    assert readme.exists()
    assert readme.read_text() == USER_README_CONTENT



"""Module for the log file and related models."""

from datetime import datetime
from enum import StrEnum
from typing import Any, Literal, Self, TypeVar

from pydantic import BaseModel, Field, RootModel

from fmu.settings.models._enums import FilterType

LogEntryType = TypeVar("LogEntryType", bound=BaseModel)


class Log(RootModel[list[LogEntryType]]):
    """Represents a log file in a .fmu directory."""

    root: list[LogEntryType] = Field(default_factory=list)

    def add_entry(self: Self, entry: LogEntryType) -> None:
        """Adds a log entry to the log."""
        self.root.append(entry)

    def __getitem__(self: Self, index: int) -> LogEntryType:
        """Retrieves a log entry from the log using the specified index."""
        return self.root[index]

    def __iter__(self: Self) -> Any:
        """Returns an iterator for the log."""
        return iter(self.root)

    def __len__(self: Self) -> int:
        """Returns the number of log entries in the log."""
        return len(self.root)


class Filter(BaseModel):
    """Represents a filter that can be applied on a log file."""

    field_name: str
    filter_value: str
    filter_type: FilterType
    operator: Literal[">=", "<=", "==", "!="]

    def parse_filter_value(self: Self) -> str | datetime | int:
        """Parse filter value to its type."""
        match self.filter_type:
            case FilterType.date:
                return datetime.fromisoformat(self.filter_value)
            case FilterType.number:
                return int(self.filter_value)
            case FilterType.text:
                return self.filter_value
            case _:
                raise ValueError(
                    "Invalid filter type supplied when parsing filter value."
                )


class LogFileName(StrEnum):
    """The log files in the .fmu directory."""

    changelog = "changelog.json"
    user_session_log = "user_session_log.json"



"""Contains the base class used for interacting with resources."""

from __future__ import annotations

import copy
import json
from typing import TYPE_CHECKING, Any, Generic, Self, TypeVar

from pydantic import BaseModel, ValidationError

from fmu.settings.models.project_config import ProjectConfig
from fmu.settings.types import ResettableBaseModel

if TYPE_CHECKING:
    # Avoid circular dependency for type hint in __init__ only
    from pathlib import Path

    from fmu.settings._fmu_dir import FMUDirectoryBase

PydanticResource = TypeVar("PydanticResource", bound=BaseModel)
MutablePydanticResource = TypeVar("MutablePydanticResource", bound=ResettableBaseModel)


class PydanticResourceManager(Generic[PydanticResource]):
    """Base class for managing resources represented by Pydantic models."""

    automatic_caching: bool = True

    def __init__(
        self: Self, fmu_dir: FMUDirectoryBase, model_class: type[PydanticResource]
    ) -> None:
        """Initializes the resource manager.

        Args:
            fmu_dir: The FMUDirectory instance
            model_class: The Pydantic model class this manager handles
        """
        self.fmu_dir = fmu_dir
        self.model_class = model_class
        self._cache: PydanticResource | None = None

    @property
    def relative_path(self: Self) -> Path:
        """Returns the path to the resource file _inside_ the .fmu directory.

        Must be implemented by subclasses.
        """
        raise NotImplementedError

    @property
    def path(self: Self) -> Path:
        """Returns the full path to the resource file."""
        return self.fmu_dir.get_file_path(self.relative_path)

    @property
    def exists(self: Self) -> bool:
        """Returns whether or not the resource exists."""
        return self.path.exists()

    @staticmethod
    def _get_dot_notation_key(
        resource_dict: dict[str, Any], key: str, default: Any = None
    ) -> Any:
        """Get a value from the resource by a dot-notation key.

        Args:
            resource_dict: The resource dictionary to get the value from
            key: The key to the value in the resource
            default: Value to return if key is not found. Default None

        Returns:
            The value or default
        """
        parts = key.split(".")
        value = resource_dict
        for part in parts:
            if isinstance(value, dict) and part in value:
                value = value[part]
            else:
                return default

        return value

    def load(
        self: Self, force: bool = False, store_cache: bool = True
    ) -> PydanticResource:
        """Loads the resource from disk and validates it as a Pydantic model.

        Args:
            force: Force a re-read even if the file is already cached.
            store_cache: Whether or not to cache the loaded model internally. This is
                best used with 'force=True' because if a model is already stored in
                _cache it will be returned without re-loading. Default True.

        Returns:
            Validated Pydantic model

        Raises:
            ValueError: If the resource file is missing or data does not match the
            model schema
        """
        if self._cache is None or force:
            if not self.exists:
                raise FileNotFoundError(
                    f"Resource file for '{self.__class__.__name__}' not found "
                    f"at: '{self.path}'"
                )

            try:
                content = self.fmu_dir.read_text_file(self.relative_path)
                data = json.loads(content)
                validated_model = self.model_class.model_validate(data)
                if store_cache:
                    self._cache = validated_model
                else:
                    return validated_model
            except ValidationError as e:
                raise ValueError(
                    f"Invalid content in resource file for '{self.__class__.__name__}: "
                    f"'{e}"
                ) from e
            except json.JSONDecodeError as e:
                raise ValueError(
                    f"Invalid JSON in resource file for '{self.__class__.__name__}': "
                    f"'{e}'"
                ) from e

        return self._cache

    def save(
        self: Self,
        model: PydanticResource,
    ) -> None:
        """Save the Pydantic model to disk.

        Args:
            model: Validated Pydantic model instance.
        """
        self.fmu_dir._lock.ensure_can_write()
        json_data = model.model_dump_json(by_alias=True, indent=2)
        self.fmu_dir.write_text_file(self.relative_path, json_data)

        if self.automatic_caching and self.exists:
            self.fmu_dir.cache.store_revision(self.relative_path, json_data)

        self._cache = model


class MutablePydanticResourceManager(PydanticResourceManager[MutablePydanticResource]):
    """Manages the .fmu resource file."""

    def __init__(
        self: Self, fmu_dir: FMUDirectoryBase, resource: type[MutablePydanticResource]
    ) -> None:
        """Initializes the resource manager."""
        super().__init__(fmu_dir, resource)

    def get(self: Self, key: str, default: Any = None) -> Any:
        """Gets a resource value by key.

        Supports dot notation for nested values (e.g., "foo.bar")

        Args:
            key: The resource key
            default: Value to return if key is not found. Default None

        Returns:
            The resource value or default
        """
        try:
            resource = self.load()

            if "." in key:
                return self._get_dot_notation_key(resource.model_dump(), key, default)

            if hasattr(resource, key):
                return getattr(resource, key)

            resource_dict = resource.model_dump()
            return resource_dict.get(key, default)
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Resource file for '{self.__class__.__name__}' not found "
                f"at: '{self.path}' when getting key {key}"
            ) from e

    def _set_dot_notation_key(
        self: Self, resource_dict: dict[str, Any], key: str, value: Any
    ) -> None:
        """Sets the value to a dot-notation key.

        Args:
            resource_dict: The resource dictionary we are modifying (by reference)
            key: The key to set
            value: The value to set
        """
        parts = key.split(".")
        target = resource_dict

        for part in parts[:-1]:
            if part not in target or not isinstance(target[part], dict):
                target[part] = {}
            target = target[part]

        target[parts[-1]] = value

    def set(self: Self, key: str, value: Any) -> None:
        """Sets a resource value by key.

        Args:
            key: The resource key
            value: The value to set

        Raises:
            FileNotFoundError: If resource file doesn't exist
            ValueError: If the updated resource is invalid
        """
        try:
            resource = self.load()
            resource_dict = resource.model_dump()
            old_resource_dict = copy.deepcopy(resource_dict)

            if "." in key:
                self._set_dot_notation_key(resource_dict, key, value)
            else:
                resource_dict[key] = value

            updated_resource = resource.model_validate(resource_dict)
            self.save(updated_resource)

            if self.model_class == ProjectConfig:
                self.fmu_dir._changelog.log_update_to_changelog(
                    updates={key: value},
                    old_resource_dict=old_resource_dict,
                    relative_path=self.relative_path,
                )

        except ValidationError as e:
            raise ValueError(
                f"Invalid value set for '{self.__class__.__name__}' with "
                f"key '{key}', value '{value}': '{e}"
            ) from e
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Resource file for '{self.__class__.__name__}' not found "
                f"at: '{self.path}' when setting key {key}"
            ) from e

    def update(self: Self, updates: dict[str, Any]) -> MutablePydanticResource:
        """Updates multiple resource values at once.

        Args:
            updates: Dictionary of key-value pairs to update

        Returns:
            The updated Resource object

        Raises:
            FileNotFoundError: If resource file doesn't exist
            ValueError: If the updates resource is invalid
        """
        try:
            resource = self.load()
            resource_dict = resource.model_dump()
            old_resource_dict = copy.deepcopy(resource_dict)

            flat_updates = {k: v for k, v in updates.items() if "." not in k}
            resource_dict.update(flat_updates)

            for key, value in updates.items():
                if "." in key:
                    self._set_dot_notation_key(resource_dict, key, value)

            updated_resource = resource.model_validate(resource_dict)
            self.save(updated_resource)

            if self.model_class == ProjectConfig:
                self.fmu_dir._changelog.log_update_to_changelog(
                    updates, old_resource_dict, self.relative_path
                )

        except ValidationError as e:
            raise ValueError(
                f"Invalid value set for '{self.__class__.__name__}' with "
                f"updates '{updates}': '{e}"
            ) from e
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Resource file for '{self.__class__.__name__}' not found "
                f"at: '{self.path}' when setting updates {updates}"
            ) from e

        return updated_resource

    def reset(self: Self) -> MutablePydanticResource:
        """Resets the resources to defaults.

        Returns:
            The new default resource object
        """
        resource = self.model_class.reset()
        self.save(resource)
        return resource



"""Model for the log entries in the the changelog file."""

from datetime import UTC, datetime
from pathlib import Path

from pydantic import AwareDatetime, BaseModel, Field

from fmu.settings.models._enums import ChangeType


class ChangeInfo(BaseModel):
    """Represents a change in the changelog file."""

    timestamp: AwareDatetime = Field(default_factory=lambda: datetime.now(UTC))
    change_type: ChangeType
    user: str
    path: Path
    change: str
    hostname: str
    file: str
    key: str



"""Tests for LogManager."""

from pathlib import Path
from typing import Self

import pytest
from pydantic import BaseModel

from fmu.settings._fmu_dir import ProjectFMUDirectory
from fmu.settings._resources.log_manager import LogManager
from fmu.settings.models._enums import FilterType
from fmu.settings.models.log import Filter, Log


def test_log_manager_instantiation(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests basic facts about the LogManager."""

    class TestEntry(BaseModel):
        user: str = "test_user"

    class TestManager(LogManager[TestEntry]):
        def __init__(self: Self, fmu_dir: ProjectFMUDirectory) -> None:
            super().__init__(fmu_dir, Log[TestEntry])

        @property
        def relative_path(self: Self) -> Path:
            return Path("logs") / "testlog.json"

    test_manager = TestManager(fmu_dir)
    assert test_manager._cached_dataframe is None
    assert test_manager.model_class == Log[TestEntry]
    with pytest.raises(
        FileNotFoundError, match="Resource file for 'TestManager' not found"
    ):
        test_manager.load()

    test_entry = TestEntry()
    test_manager.add_log_entry(test_entry)
    assert test_manager.exists
    assert test_manager.load()[0] == test_entry


def test_changelog_filtering_on_numbers(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests filtering log when filter type is a number."""

    class NumberLogEntry(BaseModel):
        count: int = 1
        user: str = "test_user"
        data: str = "some_data"

    class NumberLogManager(LogManager[NumberLogEntry]):
        def __init__(self: Self, fmu_dir: ProjectFMUDirectory) -> None:
            super().__init__(fmu_dir, Log[NumberLogEntry])

        @property
        def relative_path(self: Self) -> Path:
            return Path("logs") / "logwithnumber.json"

    first_log_entry = NumberLogEntry()
    second_log_entry = NumberLogEntry()
    second_log_entry.count = 2
    third_log_entry = NumberLogEntry()
    third_log_entry.count = 3
    log_manager = NumberLogManager(fmu_dir=fmu_dir)
    log_manager.add_log_entry(first_log_entry)
    log_manager.add_log_entry(second_log_entry)
    log_manager.add_log_entry(third_log_entry)
    assert log_manager.exists

    filter_value = 3
    filter: Filter = Filter(
        field_name="count",
        filter_value=str(filter_value),
        filter_type=FilterType.number,
        operator="==",
    )
    filtered_log = log_manager.filter_log(filter)
    assert len(filtered_log) == 1
    assert all(entry.count == filter_value for entry in filtered_log)

    filter.operator = "!="
    filtered_log = log_manager.filter_log(filter)
    expected_log_entries = 2
    assert len(filtered_log) == expected_log_entries
    assert all(entry.count != filter_value for entry in filtered_log)

    filter.operator = "<="
    filtered_log = log_manager.filter_log(filter)
    expected_log_entries = 3
    assert len(filtered_log) == expected_log_entries
    assert all(entry.count <= filter_value for entry in filtered_log)

    filter.operator = ">="
    filtered_log = log_manager.filter_log(filter)
    assert len(filtered_log) == 1
    assert all(entry.count >= filter_value for entry in filtered_log)



"""Manages an .fmu user_session_log file."""

from __future__ import annotations

import contextlib
from pathlib import Path
from typing import TYPE_CHECKING, Self

from fmu.settings._resources.log_manager import LogManager
from fmu.settings.models.event_info import EventInfo
from fmu.settings.models.log import Log, LogFileName

if TYPE_CHECKING:
    # Avoid circular dependency for type hint in __init__ only
    from fmu.settings._fmu_dir import (
        FMUDirectoryBase,
    )


class UserSessionLogManager(LogManager[EventInfo]):
    """Manages the .fmu user session log file."""

    def __init__(
        self: Self, fmu_dir: FMUDirectoryBase, retention_days: int | None = None
    ) -> None:
        """Initializes the User session log resource manager."""
        super().__init__(fmu_dir, Log[EventInfo])

        # If user_session_log.json exists from previous session, cache it and delete
        # We want a fresh log each time
        if self.exists:
            self.fmu_dir._lock.ensure_can_write()
            content = self.fmu_dir.read_text_file(self.relative_path)
            self.fmu_dir.cache.store_revision(
                self.relative_path, content, skip_trim=True
            )
            with contextlib.suppress(FileNotFoundError):
                self.path.unlink()

        self.fmu_dir.cache.trim_by_age(
            self.relative_path, retention_days or self.fmu_dir.cache.RETENTION_DAYS
        )

    @property
    def relative_path(self: Self) -> Path:
        """Returns the relative path to the log file."""
        return Path("logs") / LogFileName.user_session_log



"""Functions related to finding and validating an existing global configuration."""

from pathlib import Path
from typing import Final

from pydantic import ValidationError

from fmu.config.utilities import yaml_load
from fmu.datamodels.fmu_results.global_configuration import GlobalConfiguration

from ._logging import null_logger

logger: Final = null_logger(__name__)


class InvalidGlobalConfigurationError(ValueError):
    """Raised when a GlobalConfiguration contains invalid or disallowed content.

    This includes Drogon test data or other disallowed masterdata.
    This error is only raised when strict validation is enabled.
    """


# These should all be normalized to lower case.
INVALID_NAMES: Final[tuple[str, ...]] = (
    "drogon",
    "drogon_2020",
    "drogon_has_no_stratcolumn",
)
INVALID_UUIDS: Final[tuple[str, ...]] = (
    "ad214d85-dac7-19da-e053-c918a4889309",
    "ad214d85-8a1d-19da-e053-c918a4889310",
    "00000000-0000-0000-0000-000000000000",
)
INVALID_STRAT_NAMES: Final[tuple[str, ...]] = (
    "basevolantis",
    "basetherys",
    "basevalysar",
    "basevolon",
    "therys",
    "toptherys",
    "topvolantis",
    "topvolon",
    "topvalysar",
    "valysar",
    "volantis",
    "volon",
)


def validate_global_configuration_strictly(cfg: GlobalConfiguration) -> None:  # noqa: PLR0912
    """Does stricter checks against a valid GlobalConfiguration file.

    This is to prevent importing existing but incorrect data that should not make it
    into a project .fmu configuration. An example of this is Drogon masterdata.

    Args:
        cfg: A GlobalConfiguration instance to be validated

    Raises:
        InvalidGlobalConfigurationError: If some value in the GlobalConfiguration
            is invalid or not allowed
    """
    # Check model and access
    if cfg.model.name.lower() in INVALID_NAMES:
        raise InvalidGlobalConfigurationError(
            f"Invalid name in 'model': {cfg.model.name}"
        )
    if cfg.access.asset.name.lower() in INVALID_NAMES:
        raise InvalidGlobalConfigurationError(
            f"Invalid name in 'access.asset': {cfg.access.asset.name}"
        )

    # Check masterdata

    # smda.country
    for country in cfg.masterdata.smda.country:
        if str(country.uuid) in INVALID_UUIDS:
            raise InvalidGlobalConfigurationError(
                f"Invalid SMDA UUID in 'smda.country': {country.uuid}"
            )

    # smda.discovery
    for discovery in cfg.masterdata.smda.discovery:
        if discovery.short_identifier.lower() in INVALID_NAMES:
            raise InvalidGlobalConfigurationError(
                f"Invalid SMDA short identifier in 'smda.discovery': "
                f"{discovery.short_identifier}"
            )
        if str(discovery.uuid) in INVALID_UUIDS:
            raise InvalidGlobalConfigurationError(
                f"Invalid SMDA UUID in 'smda.discovery': {discovery.uuid}"
            )

    # smda.field
    for field in cfg.masterdata.smda.field:
        if field.identifier.lower() in INVALID_NAMES:
            raise InvalidGlobalConfigurationError(
                f"Invalid SMDA identifier in 'smda.field': {field.identifier}"
            )
        if str(field.uuid) in INVALID_UUIDS:
            raise InvalidGlobalConfigurationError(
                f"Invalid SMDA UUID in 'smda.field': {field.uuid}"
            )

    # smda.coordinate_system
    if (coord_uuid := str(cfg.masterdata.smda.coordinate_system.uuid)) in INVALID_UUIDS:
        raise InvalidGlobalConfigurationError(
            f"Invalid SMDA UUID in 'smda.coordinate_system': {coord_uuid}"
        )

    # smda.stratigraphic_column
    strat = cfg.masterdata.smda.stratigraphic_column
    if strat.identifier.lower() in INVALID_NAMES:
        raise InvalidGlobalConfigurationError(
            f"Invalid SMDA identifier in 'smda.stratigraphic_column': "
            f"{strat.identifier}"
        )
    if str(strat.uuid) in INVALID_UUIDS:
        raise InvalidGlobalConfigurationError(
            f"Invalid SMDA UUID in 'smda.stratigraphic_column': {strat.uuid}"
        )

    # Check stratigraphy

    if cfg.stratigraphy:
        for key in cfg.stratigraphy:
            if key.lower() in INVALID_STRAT_NAMES:
                raise InvalidGlobalConfigurationError(
                    f"Invalid stratigraphy name in 'smda.stratigraphy': {key}"
                )


def load_global_configuration_if_present(
    path: Path, fmu_load: bool = False
) -> GlobalConfiguration | None:
    """Loads a global config/global variables at a path.

    This loads via fmu-config, which is capable of loading a global _config_, which is
    different from the global _variables_ in that it may still be in separate files
    linked by the custom '!include' directive.

    Args:
        path: The path to the yaml file
        fmu_load: Whether or not to load in the custom 'fmu' format. Default False.

    Returns:
        GlobalConfiguration instance or None if file cannot be loaded.

    Raises:
        ValidationError: If the file is loaded but has invalid schema.
    """
    loader = "fmu" if fmu_load else "standard"
    try:
        global_variables_dict = yaml_load(path, loader=loader)
        global_config = GlobalConfiguration.model_validate(global_variables_dict)
        logger.debug(f"Global variables at {path} has valid settings data")
    except ValidationError:
        raise
    except Exception as e:
        logger.debug(
            f"Failed to load global variables at {path}: {type(e).__name__}: {e}"
        )
        return None
    return global_config


def _find_global_variables_file(paths: list[Path]) -> GlobalConfiguration | None:
    """Finds a valid global variables file, or not.

    This is the _output_ file after fmuconfig is run.

    Args:
        paths: A list of Paths to check.

    Returns:
        A validated GlobalConfiguration or None.

    Raises:
        ValidationError: If a file is found but has invalid schema.
    """
    for path in paths:
        if not path.exists():
            continue

        global_variables_path = path
        # If the path is a dir, and doesn't contain the right file, move on.
        if path.is_dir():
            global_variables_path = path / "global_variables.yml"
            if not global_variables_path.exists():
                continue

        logger.info(f"Found global variables at {path}")
        global_config = load_global_configuration_if_present(global_variables_path)
        if not global_config:
            continue
        return global_config

    return None


def _find_global_config_file(paths: list[Path]) -> GlobalConfiguration | None:
    """Finds a valid global configuration file, or not.

    This is the _input_ file, before fmuconfig is run.

    Args:
        paths: A list of Paths to check.

    Returns:
        A validated GlobalConfiguration or None.

    Raises:
        ValidationError: If a file is found but has invalid schema.
    """
    for path in paths:
        if not path.exists():
            continue

        logger.info(f"Found global config at {path}")
        # May be global_config*.yml or global_master*.yml
        for global_config_path in path.glob("**/global*.yml"):
            global_config = load_global_configuration_if_present(
                global_config_path, fmu_load=True
            )
            if not global_config:
                continue
            return global_config

    return None


def find_global_config(
    base_path: str | Path,
    extra_output_paths: list[Path] | None = None,
    extra_input_dirs: list[Path] | None = None,
    strict: bool = True,
) -> GlobalConfiguration | None:
    """Try to locate a global configuration with valid masterdata in known location.

    Extra paths may be provided

    Args:
        base_path: The path to the project root
        extra_output_paths: A list of extra paths to a global _variables_.
        extra_input_dirs: A list of extra dirs to a global _config_ might be.
        strict: If True, valid data but invalid _content_ is disallowed, i.e.  Drogon
            data. Default True.

    Returns:
        A valid GlobalConfiguration instance, or None.

    Raises:
        ValidationError: If a configuration file is found but has invalid schema.
        InvalidGlobalConfigurationError: If strict=True and configuration contains
            disallowed content (e.g., Drogon data).
    """
    base_path = Path(base_path)

    # Loads with 'fmu_load=False'
    known_output_paths = [base_path / "fmuconfig/output/global_variables.yml"]
    if extra_output_paths:
        known_output_paths += extra_output_paths

    global_config = _find_global_variables_file(known_output_paths)
    if global_config:
        if strict:
            validate_global_configuration_strictly(global_config)
        return global_config

    # Loads with 'fmu_load=True'
    known_input_paths = [base_path / "fmuconfig/input"]
    if extra_input_dirs:
        known_input_paths += extra_input_dirs

    global_config = _find_global_config_file(known_input_paths)
    if global_config:
        if strict:
            validate_global_configuration_strictly(global_config)
        return global_config

    logger.info("No global variables or config with valid settings data found.")
    return None



"""Tests for UserSessionLogManager."""

import os
from datetime import UTC, datetime, timedelta
from pathlib import Path

from fmu.settings._fmu_dir import ProjectFMUDirectory
from fmu.settings._resources.cache_manager import CacheManager
from fmu.settings._resources.user_session_log_manager import UserSessionLogManager
from fmu.settings.models.event_info import EventInfo


def test_user_session_log_manager_creates_fresh_log(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that UserSessionLogManager starts with no existing log file."""
    manager = UserSessionLogManager(fmu_dir)
    assert not manager.exists
    assert manager.relative_path == Path("logs") / "user_session_log.json"


def test_user_session_log_manager_add_log_entry(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that UserSessionLogManager can add log entries."""
    manager = UserSessionLogManager(fmu_dir)
    event = EventInfo(level="INFO", event="test_event")
    manager.add_log_entry(event)

    assert manager.exists
    log_entries = manager.load()
    assert len(log_entries) == 1
    assert log_entries[0].event == "test_event"
    assert log_entries[0].level == "INFO"


def test_user_session_log_manager_archives_previous_session(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that previous session log is cached on initialization."""
    manager1 = UserSessionLogManager(fmu_dir)
    event1 = EventInfo(level="INFO", event="session1_event")
    manager1.add_log_entry(event1)
    assert manager1.exists

    manager2 = UserSessionLogManager(fmu_dir)
    assert not manager2.exists

    cache_path = fmu_dir.path / "cache" / "user_session_log"
    cached_files = list(cache_path.glob("*.json"))
    assert len(cached_files) == 1

    event2 = EventInfo(level="INFO", event="session2_event")
    manager2.add_log_entry(event2)
    log_entries = manager2.load()
    assert len(log_entries) == 1
    assert log_entries[0].event == "session2_event"


def test_user_session_log_manager_trims_old_cache_on_init(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that old cached logs are trimmed with default and custom retention."""
    cache_dir = fmu_dir.path / "cache" / "user_session_log"
    cache_dir.mkdir(parents=True, exist_ok=True)

    now = datetime.now(UTC)
    very_old_time = now - timedelta(days=CacheManager.RETENTION_DAYS + 5)
    old_time = now - timedelta(days=15)
    recent_time = now - timedelta(days=5)

    very_old_filename = very_old_time.strftime("%Y%m%dT%H%M%S.%fZ") + "-very_old.json"
    old_filename = old_time.strftime("%Y%m%dT%H%M%S.%fZ") + "-old.json"
    recent_filename = recent_time.strftime("%Y%m%dT%H%M%S.%fZ") + "-recent.json"

    very_old_file = cache_dir / very_old_filename
    old_file = cache_dir / old_filename
    recent_file = cache_dir / recent_filename
    very_old_file.write_text('{"entries": []}', encoding="utf-8")
    old_file.write_text('{"entries": []}', encoding="utf-8")
    recent_file.write_text('{"entries": []}', encoding="utf-8")

    very_old_mtime = very_old_time.timestamp()
    old_mtime = old_time.timestamp()
    recent_mtime = recent_time.timestamp()
    os.utime(very_old_file, (very_old_mtime, very_old_mtime))
    os.utime(old_file, (old_mtime, old_mtime))
    os.utime(recent_file, (recent_mtime, recent_mtime))

    UserSessionLogManager(fmu_dir)

    remaining_files = sorted(cache_dir.glob("*.json"), key=lambda p: p.name)
    assert len(remaining_files) == 2  # noqa: PLR2004
    assert remaining_files[0].name == old_filename
    assert remaining_files[1].name == recent_filename

    UserSessionLogManager(fmu_dir, retention_days=10)

    remaining_files = list(cache_dir.glob("*.json"))
    assert len(remaining_files) == 1
    assert remaining_files[0].name == recent_filename


def test_user_session_log_manager_skip_trim_on_store_revision(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests that UserSessionLogManager uses skip_trim=True when storing revisions."""
    for i in range(10):
        manager = UserSessionLogManager(fmu_dir)
        event = EventInfo(level="INFO", event=f"session_{i}_event")
        manager.add_log_entry(event)

    cache_dir = fmu_dir.path / "cache" / "user_session_log"
    cached_files = list(cache_dir.glob("*.json"))
    assert len(cached_files) == 9  # noqa: PLR2004


def test_user_session_log_manager_multiple_entries_in_session(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests handling multiple log entries in one session."""
    manager = UserSessionLogManager(fmu_dir)

    events = [
        EventInfo(level="INFO", event="event1"),
        EventInfo(level="WARNING", event="event2"),
        EventInfo(level="ERROR", event="event3"),
    ]

    for event in events:
        manager.add_log_entry(event)

    log_entries = manager.load()
    assert len(log_entries) == 3  # noqa: PLR2004
    assert [entry.event for entry in log_entries] == ["event1", "event2", "event3"]
    assert [entry.level for entry in log_entries] == ["INFO", "WARNING", "ERROR"]



"""Model for the log entries in the the userlog file."""

from datetime import UTC, datetime

from pydantic import AwareDatetime, BaseModel, ConfigDict, Field


class EventInfo(BaseModel):
    """Represents information about user session activity."""

    model_config = ConfigDict(extra="allow")

    level: str = "INFO"
    event: str = "unknown"
    timestamp: AwareDatetime = Field(default_factory=lambda: datetime.now(UTC))



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Self

import pandas as pd
from pydantic import ValidationError

from fmu.settings._resources.pydantic_resource_manager import PydanticResourceManager
from fmu.settings.models._enums import FilterType
from fmu.settings.models.log import Filter, Log, LogEntryType

if TYPE_CHECKING:
    # Avoid circular dependency for type hint in __init__ only
    from fmu.settings._fmu_dir import (
        FMUDirectoryBase,
    )


class LogManager(PydanticResourceManager[Log[LogEntryType]], Generic[LogEntryType]):
    """Manages the .fmu log files."""

    automatic_caching: bool = False

    def __init__(
        self: Self, fmu_dir: FMUDirectoryBase, model_class: type[Log[LogEntryType]]
    ) -> None:
        """Initializes the log resource manager."""
        self._cached_dataframe: pd.DataFrame | None = None
        super().__init__(fmu_dir, model_class)

    def add_log_entry(self: Self, log_entry: LogEntryType) -> None:
        """Adds a log entry to the log resource."""
        try:
            validated_entry = log_entry.model_validate(log_entry.model_dump())
            log_model: Log[LogEntryType] = (
                self.load() if self.exists else self.model_class([])
            )
            log_model.add_entry(validated_entry)
            self.save(log_model)
            self._cached_dataframe = None
        except ValidationError as e:
            raise ValueError(
                f"Invalid log entry added to '{self.model_class.__name__}' with "
                f"value '{log_entry}': '{e}"
            ) from e

    def filter_log(self: Self, filter: Filter) -> Log[LogEntryType]:
        """Filters the log resource with the provided filter."""
        if self._cached_dataframe is None:
            log_model: Log[LogEntryType] = self.load()
            df_log = pd.DataFrame([entry.model_dump() for entry in log_model])
            self._cached_dataframe = df_log
        df_log = self._cached_dataframe

        match filter.operator:
            case "==":
                filtered_df = df_log[
                    df_log[filter.field_name] == filter.parse_filter_value()
                ]
            case "!=":
                filtered_df = df_log[
                    df_log[filter.field_name] != filter.parse_filter_value()
                ]
            case "<=":
                if filter.filter_type == FilterType.text:
                    raise ValueError(
                        f"Invalid filter operator <= applied to '{FilterType.text}' "
                        f"field {filter.field_name} when filtering log resource "
                        f"{self.model_class.__name__} with value {filter.filter_value}."
                    )
                filtered_df = df_log[
                    df_log[filter.field_name] <= filter.parse_filter_value()
                ]
            case ">=":
                if filter.filter_type == FilterType.text:
                    raise ValueError(
                        f"Invalid filter operator >= applied to '{FilterType.text}' "
                        f"field {filter.field_name} when filtering log resource "
                        f"{self.model_class.__name__} with value {filter.filter_value}."
                    )
                filtered_df = df_log[
                    df_log[filter.field_name] >= filter.parse_filter_value()
                ]
            case _:
                raise ValueError(
                    "Invalid filter operator applied when "
                    f"filtering log resource {self.model_class.__name__} "
                )

        filtered_dict = filtered_df.to_dict("records")
        return self.model_class.model_validate(filtered_dict)



"""Tests for fmu.resources.config."""

import copy
import json
from datetime import datetime
from pathlib import Path
from typing import Any

import pytest
from fmu.datamodels.fmu_results.fields import Access, Model, Smda

from fmu.settings._fmu_dir import ProjectFMUDirectory, UserFMUDirectory
from fmu.settings._resources.config_managers import (
    ProjectConfigManager,
    UserConfigManager,
)
from fmu.settings.models._enums import ChangeType
from fmu.settings.models.project_config import ProjectConfig
from fmu.settings.models.user_config import UserConfig


@pytest.fixture
def nested_dict() -> dict[str, Any]:
    """A nested dictionary to test get-set dot notation."""
    return {
        "a": 1,
        "b": {
            "c": "2",
            "d": {"e": 3},
        },
    }


def test_config_resource_manager(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests basic facts about the ProjectConfig resource manager."""
    # This manager already exists in 'fmu_dir', but just to
    # try from scratch.
    manager = ProjectConfigManager(fmu_dir)

    assert manager.fmu_dir == fmu_dir
    assert manager.model_class == ProjectConfig
    assert manager._cache is None
    # Resource manager requires this to be implemented
    assert manager.relative_path == Path("config.json")
    assert manager.path == fmu_dir.path / manager.relative_path


def test_user_config_resource_manager(user_fmu_dir: UserFMUDirectory) -> None:
    """Tests basic facts about the ProjectConfig resource manager."""
    # This manager already exists in 'fmu_dir', but just to
    # try from scratch.
    manager = UserConfigManager(user_fmu_dir)

    assert manager.fmu_dir == user_fmu_dir
    assert manager.model_class == UserConfig
    assert manager._cache is None
    # Resource manager requires this to be implemented
    assert manager.relative_path == Path("config.json")
    assert manager.path == user_fmu_dir.path / manager.relative_path


def test_load_config(
    fmu_dir: ProjectFMUDirectory,
    config_model: ProjectConfig,
) -> None:
    """Tests that load() works when a configuration exists."""
    assert fmu_dir.config.load() == config_model
    assert fmu_dir.config._cache == config_model


def test_load_user_config(
    user_fmu_dir: UserFMUDirectory,
    user_config_model: UserConfig,
) -> None:
    """Tests that load() works when a user configuration exists."""
    assert user_fmu_dir.config.load() == user_config_model
    assert user_fmu_dir.config._cache == user_config_model


def test_get_config_missing_file(tmp_path: Path) -> None:
    """Tests get_config raises FileNotFoundError when config.json missing."""
    empty_fmu_dir = tmp_path / ".fmu"
    empty_fmu_dir.mkdir()

    fmu_dir = ProjectFMUDirectory(tmp_path)

    assert fmu_dir.config.exists is False
    with pytest.raises(
        FileNotFoundError,
        match=(
            "Resource file for 'ProjectConfigManager' not found at: "
            f"'{empty_fmu_dir}/config.json'"
        ),
    ):
        fmu_dir.config.load()


def test_get_config_invalid_json(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests a corrupted config.json raises ValueError."""
    config_path = fmu_dir.path / "config.json"
    with open(config_path, "a", encoding="utf-8") as f:
        f.write("%")

    with pytest.raises(
        ValueError, match="Invalid JSON in resource file for 'ProjectConfigManager'"
    ):
        fmu_dir.config.load(force=True)


def test_get_dot_notation_key(
    nested_dict: dict[str, Any], fmu_dir: ProjectFMUDirectory
) -> None:
    """Tests the get helper function for dot notation works as expected."""
    assert fmu_dir.config._get_dot_notation_key(nested_dict, "b.c") == "2"
    assert fmu_dir.config._get_dot_notation_key(nested_dict, "b.d") == {"e": 3}
    assert fmu_dir.config._get_dot_notation_key(nested_dict, "b.d.e") == 3  # noqa PLR2004
    assert fmu_dir.config._get_dot_notation_key(nested_dict, "b.z") is None
    assert fmu_dir.config._get_dot_notation_key(nested_dict, "b.z", "foo") == "foo"


def test_get_key(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests getting a key on configuration."""
    assert fmu_dir.config.get("created_by") == "user"
    # No nested entries exist in config yet.
    assert fmu_dir.config.get("does.not.exist", "foo") == "foo"
    assert fmu_dir.config.get("notreal", "foo") == "foo"


def test_get_key_config_does_not_exist(tmp_path: Path) -> None:
    """Tests getting a key when the config is missing."""
    empty_fmu_dir = tmp_path / ".fmu"
    empty_fmu_dir.mkdir()

    fmu_dir = ProjectFMUDirectory(tmp_path)

    assert fmu_dir.config.exists is False
    with pytest.raises(
        FileNotFoundError,
        match=(
            "Resource file for 'ProjectConfigManager' not found at: "
            f"'{empty_fmu_dir}/config.json'"
        ),
    ):
        fmu_dir.config.get("version")


def test_set_dot_notation_key(
    nested_dict: dict[str, Any], fmu_dir: ProjectFMUDirectory
) -> None:
    """Tests the set helper function for dot notation works as expected."""
    assert nested_dict["b"]["c"] == "2"
    fmu_dir.config._set_dot_notation_key(nested_dict, "b.c", "3")
    assert nested_dict["b"]["c"] == "3"

    assert nested_dict["b"]["d"] == {"e": 3}
    fmu_dir.config._set_dot_notation_key(nested_dict, "b.d", {"e": "s"})
    assert nested_dict["b"]["d"] == {"e": "s"}
    fmu_dir.config._set_dot_notation_key(nested_dict, "b.d.e", 3)
    assert nested_dict["b"]["d"] == {"e": 3}

    assert fmu_dir.config._get_dot_notation_key(nested_dict, "b.z") is None
    fmu_dir.config._set_dot_notation_key(nested_dict, "b.z", "foo")
    assert nested_dict["b"]["z"] == "foo"

    # Overwrites existing non-dict values
    fmu_dir.config._set_dot_notation_key(nested_dict, "a.b", {})
    assert nested_dict["a"] != 1
    assert nested_dict["a"]["b"] == {}


def test_set_key(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests setting a key on configuration."""
    fmu_dir.config.set("created_by", "user2")
    assert fmu_dir.config.get("created_by") == "user2"

    config_model = fmu_dir.config.load(force=True)
    config_dict = config_model.model_dump()
    assert config_dict["created_by"] == "user2"

    # No nested entries exist in config yet.
    fmu_dir.config.set("does.not.exist", "foo")
    assert fmu_dir.config.get("does.not.exist") is None
    fmu_dir.config.set("notreal", "foo")
    assert fmu_dir.config.get("notreal") is None

    # Does not write invalid values
    config_model = fmu_dir.config.load(force=True)
    config_dict = config_model.model_dump()
    assert config_dict.get("does", None) is None
    assert config_dict.get("notreal", None) is None

    with pytest.raises(
        ValueError,
        match=(
            "Invalid value set for 'ProjectConfigManager' with key 'version', "
            "value '2.0'"
        ),
    ):
        fmu_dir.config.set("version", 2.0)


def test_set_key_config_does_not_exist(tmp_path: Path) -> None:
    """Tests getting a key when the config is missing."""
    empty_fmu_dir = tmp_path / ".fmu"
    empty_fmu_dir.mkdir()

    fmu_dir = ProjectFMUDirectory(tmp_path)

    assert fmu_dir.config.exists is False
    with pytest.raises(
        FileNotFoundError,
        match=(
            "Resource file for 'ProjectConfigManager' not found at: "
            f"'{empty_fmu_dir}/config.json'"
        ),
    ):
        fmu_dir.config.set("version", "200.0.0")


def test_update_config(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests setting a key on configuration."""
    fmu_dir.config.update({"created_by": "user2", "version": "200.0.0", "not.real": 0})
    assert fmu_dir.config.get("created_by") == "user2"
    assert fmu_dir.config.get("version") == "200.0.0"
    assert fmu_dir.config.get("not") is None

    bad_updates = {"created_by": {}, "version": "major"}
    with pytest.raises(
        ValueError,
        match="Invalid value set for 'ProjectConfigManager' with updates "
        f"'{bad_updates}'",
    ):
        fmu_dir.config.update(bad_updates)


def test_update_config_writes_to_changelog(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that config updates are written to changelog."""
    fmu_dir.config.update({"created_by": "user2", "version": "200.0.0", "new.field": 0})
    changelog = fmu_dir._changelog.load()
    expected_log_entries = 3
    assert len(changelog) == expected_log_entries
    assert changelog[0].change_type == ChangeType.update
    assert changelog[0].key == "created_by"
    assert changelog[0].file == "config.json"
    assert changelog[2].change_type == ChangeType.add
    assert changelog[2].key == "new.field"
    assert changelog[2].file == "config.json"


def test_set_smda(
    fmu_dir: ProjectFMUDirectory, masterdata_dict: dict[str, Any]
) -> None:
    """Tests setting the masterdata.smda value in the config."""
    assert fmu_dir.config.get("masterdata.smda") is None
    with open(fmu_dir.path / fmu_dir.config.relative_path, encoding="utf-8") as f:
        config_on_disk = json.loads(f.read())
    assert config_on_disk["masterdata"] is None

    fmu_dir.set_config_value("masterdata", masterdata_dict)

    smda_model = Smda.model_validate(masterdata_dict["smda"])
    # Compare to round-trip'd Pydantic model for uuid-str to UUID() obj conversion
    assert fmu_dir.get_config_value("masterdata.smda") == smda_model.model_dump()
    with open(fmu_dir.path / fmu_dir.config.relative_path, encoding="utf-8") as f:
        config_on_disk = json.loads(f.read())

    config_on_disk_model = ProjectConfig.model_validate(config_on_disk)
    assert fmu_dir.config._cache is not None
    assert fmu_dir.config._cache.masterdata is not None
    assert fmu_dir.config._cache.masterdata.smda == smda_model
    assert config_on_disk_model == fmu_dir.config._cache


def test_set_config_writes_to_changelog(
    fmu_dir: ProjectFMUDirectory, masterdata_dict: dict[str, Any]
) -> None:
    """Tests setting a config value writes update to changelog."""
    fmu_dir.set_config_value("masterdata", masterdata_dict)
    changelog = fmu_dir._changelog.load()
    assert changelog[0].key == "masterdata"
    assert changelog[0].change_type == ChangeType.update
    assert changelog[0].file == "config.json"

    field = fmu_dir.get_config_value("masterdata.smda.field")
    new_field = copy.deepcopy(field)
    new_field[0]["identifier"] = "test"

    fmu_dir.set_config_value("masterdata.smda.field", new_field)
    changelog = fmu_dir._changelog.load()
    expected_log_entries = 2
    assert len(changelog) == expected_log_entries
    assert changelog[1].key == "masterdata.smda.field"
    assert changelog[1].change_type == ChangeType.update
    assert changelog[1].file == "config.json"


def test_set_model_invalid_fails(
    fmu_dir: ProjectFMUDirectory, model_dict: dict[str, Any]
) -> None:
    """Tests setting the model value in the config using an invalid dictionary."""
    assert fmu_dir.config.get("model") is None

    # drop model.name to test validation
    model_dict.pop("name")

    with pytest.raises(ValueError, match="model.name"):
        fmu_dir.set_config_value("model", model_dict)


def test_set_model(fmu_dir: ProjectFMUDirectory, model_dict: dict[str, Any]) -> None:
    """Tests setting the model value in the config."""
    assert fmu_dir.config.get("model") is None
    with open(fmu_dir.path / fmu_dir.config.relative_path, encoding="utf-8") as f:
        config_on_disk = json.loads(f.read())
    assert config_on_disk["model"] is None

    fmu_dir.set_config_value("model", model_dict)

    model = Model.model_validate(model_dict)

    assert fmu_dir.get_config_value("model") == model
    assert fmu_dir.get_config_value("model.revision") == "21.0.0"
    assert fmu_dir.get_config_value("model.name") == "Drogon"

    with open(fmu_dir.path / fmu_dir.config.relative_path, encoding="utf-8") as f:
        config_on_disk = json.loads(f.read())

    config_on_disk_model = ProjectConfig.model_validate(config_on_disk)
    assert fmu_dir.config._cache is not None
    assert fmu_dir.config._cache.model == model
    assert config_on_disk_model == fmu_dir.config._cache


def test_set_access_invalid_fails(
    fmu_dir: ProjectFMUDirectory, access_dict: dict[str, Any]
) -> None:
    """Tests setting the access value in the config using an invalid dictionary."""
    assert fmu_dir.config.get("access") is None

    # drop access.asset to test validation
    access_dict.pop("asset")

    with pytest.raises(ValueError, match="access.asset"):
        fmu_dir.set_config_value("access", access_dict)


def test_set_access(fmu_dir: ProjectFMUDirectory, access_dict: dict[str, Any]) -> None:
    """Tests setting the access value in the config."""
    assert fmu_dir.config.get("access") is None
    with open(fmu_dir.path / fmu_dir.config.relative_path, encoding="utf-8") as f:
        config_on_disk = json.loads(f.read())
    assert config_on_disk["access"] is None

    fmu_dir.set_config_value("access", access_dict)

    access = Access.model_validate(access_dict)

    assert fmu_dir.get_config_value("access") == access
    assert fmu_dir.get_config_value("access.classification") == "internal"
    assert fmu_dir.get_config_value("access.asset.name") == "Drogon"

    with open(fmu_dir.path / fmu_dir.config.relative_path, encoding="utf-8") as f:
        config_on_disk = json.loads(f.read())

    config_on_disk_model = ProjectConfig.model_validate(config_on_disk)
    assert fmu_dir.config._cache is not None
    assert fmu_dir.config._cache.access == access
    assert config_on_disk_model == fmu_dir.config._cache


def test_update_config_when_it_does_not_exist(tmp_path: Path) -> None:
    """Tests getting a key when the config is missing."""
    empty_fmu_dir = tmp_path / ".fmu"
    empty_fmu_dir.mkdir()

    fmu_dir = ProjectFMUDirectory(tmp_path)

    assert fmu_dir.config.exists is False
    with pytest.raises(
        FileNotFoundError,
        match=(
            "Resource file for 'ProjectConfigManager' not found at: "
            f"'{empty_fmu_dir}/config.json'"
        ),
    ):
        fmu_dir.config.update({"created_by": "user", "version": "200.0.0"})


def test_save(fmu_dir: ProjectFMUDirectory, config_dict: dict[str, Any]) -> None:
    """Tests that save functions as expected."""
    config = fmu_dir.config.load()
    assert fmu_dir.config._cache == config
    assert config.created_by == "user"

    config_dict["created_by"] = "user2"
    new_config = ProjectConfig.model_validate(config_dict)
    fmu_dir.config.save(new_config)

    assert fmu_dir.config._cache == new_config
    with open(fmu_dir.config.path, encoding="utf-8") as f:
        new_config_dict = json.loads(f.read())

    # Patch the raw datetime string from json
    new_config_dict["created_at"] = datetime.fromisoformat(
        new_config_dict["created_at"]
    )
    assert config_dict == new_config_dict



"""Tests for fmu.settings._global_config."""

import shutil
from collections.abc import Callable
from pathlib import Path
from typing import Any
from uuid import UUID, uuid4

import pytest
import yaml
from fmu.datamodels.fmu_results import fields
from fmu.datamodels.fmu_results.global_configuration import (
    GlobalConfiguration,
    StratigraphyElement,
)
from pydantic import ValidationError

from fmu.settings._global_config import (
    InvalidGlobalConfigurationError,
    _find_global_config_file,
    _find_global_variables_file,
    find_global_config,
    load_global_configuration_if_present,
    validate_global_configuration_strictly,
)

STRICT_VALIDATION_IDENTIFIERS: list[tuple[str, bool]] = [
    ("Drogon", False),  # False = not valid, True = valid
    ("DROGON", False),
    ("drogon_2020", False),
    ("Dragon", True),
    ("TROLL", True),
]

STRICT_VALIDATION_UUIDS: list[tuple[UUID, bool]] = [
    (UUID("ad214d85-dac7-19da-e053-c918a4889309"), False),
    (UUID("ad214d85-8a1d-19da-e053-c918a4889310"), False),
    (UUID("00000000-0000-0000-0000-000000000000"), False),
    (uuid4(), True),
]

STRICT_VALIDATION_STRAT_COLS: list[tuple[str, bool]] = [
    ("TopVolantis", False),
    ("Valysar", False),
    ("Viking", True),
]


# Test validation of invalid content/values.


@pytest.mark.parametrize("name, valid", STRICT_VALIDATION_IDENTIFIERS)
def test_validate_global_config_strict_model(
    name: str,
    valid: bool,
    generate_strict_valid_globalconfiguration: Callable[[], GlobalConfiguration],
) -> None:
    """Tests strict validation on 'model'."""
    cfg = generate_strict_valid_globalconfiguration(
        model=fields.Model(name=name, revision=""),  # type: ignore
    )
    if valid:
        validate_global_configuration_strictly(cfg)  # Does not raise
    else:
        with pytest.raises(
            InvalidGlobalConfigurationError, match=f"Invalid name in 'model': {name}"
        ):
            validate_global_configuration_strictly(cfg)


@pytest.mark.parametrize("name, valid", STRICT_VALIDATION_IDENTIFIERS)
def test_validate_global_config_strict_access(
    name: str,
    valid: bool,
    generate_strict_valid_globalconfiguration: Callable[[], GlobalConfiguration],
) -> None:
    """Tests strict validation on 'access'."""
    cfg = generate_strict_valid_globalconfiguration(
        asset=fields.Asset(name=name),  # type: ignore
    )
    if valid:
        validate_global_configuration_strictly(cfg)  # Does not raise
    else:
        with pytest.raises(
            InvalidGlobalConfigurationError,
            match=f"Invalid name in 'access.asset': {name}",
        ):
            validate_global_configuration_strictly(cfg)


@pytest.mark.parametrize("uuid, valid", STRICT_VALIDATION_UUIDS)
def test_validate_global_config_strict_smda_country_uuid(
    uuid: UUID,
    valid: bool,
    generate_strict_valid_globalconfiguration: Callable[[], GlobalConfiguration],
) -> None:
    """Tests strict validation on 'smda.country' uuids."""
    cfg = generate_strict_valid_globalconfiguration(
        country_items=[  # type: ignore
            fields.CountryItem(identifier="bar", uuid=uuid),
            fields.CountryItem(identifier="foo", uuid=uuid4()),
        ],
    )
    if valid:
        validate_global_configuration_strictly(cfg)  # Does not raise
    else:
        with pytest.raises(
            InvalidGlobalConfigurationError,
            match=f"Invalid SMDA UUID in 'smda.country': {uuid}",
        ):
            validate_global_configuration_strictly(cfg)


@pytest.mark.parametrize("identifier, valid", STRICT_VALIDATION_IDENTIFIERS)
def test_validate_global_config_strict_smda_discovery_identifier(
    identifier: str,
    valid: bool,
    generate_strict_valid_globalconfiguration: Callable[[], GlobalConfiguration],
) -> None:
    """Tests strict validation on 'smda.discovery' identifiers."""
    cfg = generate_strict_valid_globalconfiguration(
        discovery_items=[  # type: ignore
            fields.DiscoveryItem(short_identifier=identifier, uuid=uuid4()),
            fields.DiscoveryItem(short_identifier="foo", uuid=uuid4()),
        ],
    )
    if valid:
        validate_global_configuration_strictly(cfg)  # Does not raise
    else:
        with pytest.raises(
            InvalidGlobalConfigurationError,
            match=f"Invalid SMDA short identifier in 'smda.discovery': {identifier}",
        ):
            validate_global_configuration_strictly(cfg)


@pytest.mark.parametrize("uuid, valid", STRICT_VALIDATION_UUIDS)
def test_validate_global_config_strict_smda_discovery_uuid(
    uuid: UUID,
    valid: bool,
    generate_strict_valid_globalconfiguration: Callable[[], GlobalConfiguration],
) -> None:
    """Tests strict validation on 'smda.discovery' uuids."""
    cfg = generate_strict_valid_globalconfiguration(
        discovery_items=[  # type: ignore
            fields.DiscoveryItem(short_identifier="bar", uuid=uuid),
            fields.DiscoveryItem(short_identifier="foo", uuid=uuid4()),
        ],
    )
    if valid:
        validate_global_configuration_strictly(cfg)  # Does not raise
    else:
        with pytest.raises(
            InvalidGlobalConfigurationError,
            match=f"Invalid SMDA UUID in 'smda.discovery': {uuid}",
        ):
            validate_global_configuration_strictly(cfg)


@pytest.mark.parametrize("identifier, valid", STRICT_VALIDATION_IDENTIFIERS)
def test_validate_global_config_strict_smda_field_identifier(
    identifier: str,
    valid: bool,
    generate_strict_valid_globalconfiguration: Callable[[], GlobalConfiguration],
) -> None:
    """Tests strict validation on 'smda.discovery' identifiers."""
    cfg = generate_strict_valid_globalconfiguration(
        field_items=[  # type: ignore
            fields.FieldItem(identifier=identifier, uuid=uuid4()),
            fields.FieldItem(identifier="foo", uuid=uuid4()),
        ],
    )
    if valid:
        validate_global_configuration_strictly(cfg)  # Does not raise
    else:
        with pytest.raises(
            InvalidGlobalConfigurationError,
            match=f"Invalid SMDA identifier in 'smda.field': {identifier}",
        ):
            validate_global_configuration_strictly(cfg)


@pytest.mark.parametrize("uuid, valid", STRICT_VALIDATION_UUIDS)
def test_validate_global_config_strict_smda_field_uuid(
    uuid: UUID,
    valid: bool,
    generate_strict_valid_globalconfiguration: Callable[[], GlobalConfiguration],
) -> None:
    """Tests strict validation on 'smda.discovery' uuids."""
    cfg = generate_strict_valid_globalconfiguration(
        field_items=[  # type: ignore
            fields.FieldItem(identifier="bar", uuid=uuid),
            fields.FieldItem(identifier="foo", uuid=uuid4()),
        ],
    )
    if valid:
        validate_global_configuration_strictly(cfg)  # Does not raise
    else:
        with pytest.raises(
            InvalidGlobalConfigurationError,
            match=f"Invalid SMDA UUID in 'smda.field': {uuid}",
        ):
            validate_global_configuration_strictly(cfg)


@pytest.mark.parametrize("uuid, valid", STRICT_VALIDATION_UUIDS)
def test_validate_global_config_strict_coordinate_system(
    uuid: UUID,
    valid: bool,
    generate_strict_valid_globalconfiguration: Callable[[], GlobalConfiguration],
) -> None:
    """Tests strict validation on 'smda.coordinate_system'."""
    cfg = generate_strict_valid_globalconfiguration(
        coordinate_system=fields.CoordinateSystem(identifier="", uuid=uuid),  # type: ignore
    )
    if valid:
        validate_global_configuration_strictly(cfg)  # Does not raise
    else:
        with pytest.raises(
            InvalidGlobalConfigurationError,
            match=f"Invalid SMDA UUID in 'smda.coordinate_system': {uuid}",
        ):
            validate_global_configuration_strictly(cfg)


@pytest.mark.parametrize("uuid, valid", STRICT_VALIDATION_UUIDS)
def test_validate_global_config_strict_stratigraphic_column_uuids(
    uuid: UUID,
    valid: bool,
    generate_strict_valid_globalconfiguration: Callable[[], GlobalConfiguration],
) -> None:
    """Tests strict validation on 'smda.stratigraphic_column' uuid."""
    cfg = generate_strict_valid_globalconfiguration(
        stratigraphic_column=fields.StratigraphicColumn(identifier="", uuid=uuid),  # type: ignore
    )
    if valid:
        validate_global_configuration_strictly(cfg)  # Does not raise
    else:
        with pytest.raises(
            InvalidGlobalConfigurationError,
            match=f"Invalid SMDA UUID in 'smda.stratigraphic_column': {uuid}",
        ):
            validate_global_configuration_strictly(cfg)


@pytest.mark.parametrize("identifier, valid", STRICT_VALIDATION_IDENTIFIERS)
def test_validate_global_config_strict_stratigraphic_column_names(
    identifier: str,
    valid: bool,
    generate_strict_valid_globalconfiguration: Callable[[], GlobalConfiguration],
) -> None:
    """Tests strict validation on 'smda.stratigraphic_column' identifiers."""
    cfg = generate_strict_valid_globalconfiguration(
        stratigraphic_column=fields.StratigraphicColumn(  # type: ignore
            identifier=identifier, uuid=uuid4()
        ),
    )
    if valid:
        validate_global_configuration_strictly(cfg)  # Does not raise
    else:
        with pytest.raises(
            InvalidGlobalConfigurationError,
            match=f"Invalid SMDA identifier in 'smda.stratigraphic_column': "
            f"{identifier}",
        ):
            validate_global_configuration_strictly(cfg)


@pytest.mark.parametrize("identifier, valid", STRICT_VALIDATION_STRAT_COLS)
def test_validate_global_config_strict_stratigraphy_names(
    identifier: str,
    valid: bool,
    generate_strict_valid_globalconfiguration: Callable[[], GlobalConfiguration],
) -> None:
    """Tests strict validation on 'smda.stratigraphic_column' identifiers."""
    cfg = generate_strict_valid_globalconfiguration()
    assert cfg.stratigraphy
    cfg.stratigraphy.root[identifier] = StratigraphyElement(
        name=identifier, stratigraphic=False
    )
    if valid:
        validate_global_configuration_strictly(cfg)  # Does not raise
    else:
        with pytest.raises(
            InvalidGlobalConfigurationError,
            match=f"Invalid stratigraphy name in 'smda.stratigraphy': {identifier}",
        ):
            validate_global_configuration_strictly(cfg)


# Test load_global_configuration_if_present


@pytest.mark.parametrize("fmu_load", [True, False])
def test_load_global_configuration_raises_on_invalid_yaml_structure(
    fmu_load: bool, tmp_path: Path
) -> None:
    """Tests that ValidationError is raised for invalid YAML structure."""
    config_path = tmp_path / "global_config.yml"
    with open(config_path, "w") as f:
        f.write("foo=bar")

    with pytest.raises(ValidationError):
        load_global_configuration_if_present(config_path, fmu_load=fmu_load)


@pytest.mark.parametrize("fmu_load", [True, False])
def test_load_global_configuration_raises_on_missing_required_fields(
    fmu_load: bool,
    tmp_path: Path,
    global_variables_with_masterdata: dict[str, Any],
) -> None:
    """Tests that ValidationError is raised for missing required fields."""
    config_path = tmp_path / "global_config.yml"
    del global_variables_with_masterdata["masterdata"]
    with open(config_path, "w") as f:
        yaml.safe_dump(global_variables_with_masterdata, f)

    with pytest.raises(ValidationError):
        load_global_configuration_if_present(config_path, fmu_load=fmu_load)


@pytest.mark.parametrize("fmu_load", [True, False])
def test_load_global_configuration_returns_none_on_file_not_found(
    fmu_load: bool, tmp_path: Path
) -> None:
    """Tests that None is returned when file doesn't exist."""
    config_path = tmp_path / "non_existent_file.yml"
    assert load_global_configuration_if_present(config_path, fmu_load=fmu_load) is None


@pytest.mark.parametrize("fmu_load", [True, False])
def test_load_global_configuration_returns_none_on_yaml_parse_error(
    fmu_load: bool, tmp_path: Path
) -> None:
    """Tests that None is returned on YAML parsing errors."""
    config_path = tmp_path / "invalid.yml"
    with open(config_path, "w") as f:
        f.write("key: [unclosed list")
    assert load_global_configuration_if_present(config_path, fmu_load=fmu_load) is None


# Test finding global variables


def test_find_global_config_file_not_there(tmp_path: Path) -> None:
    """Tests finding the global config file if it is not present."""
    assert _find_global_config_file([]) is None
    assert _find_global_config_file([tmp_path]) is None
    assert _find_global_config_file([tmp_path / "dne"]) is None


def test_find_global_config_file_malformed_raises_validation_error(
    tmp_path: Path,
) -> None:
    """Tests that malformed global config file raises ValidationError."""
    with open(tmp_path / "global_master_config.yml", "w") as f:
        f.write("foo: bar")
    with pytest.raises(ValidationError):
        _find_global_config_file([tmp_path])


def test_find_global_config_file_skips_invalid_yaml_and_continues(
    tmp_path: Path,
) -> None:
    """Tests that function skips files with YAML parse errors and continues."""
    with open(tmp_path / "global_config.yml", "w") as f:
        f.write("key: [unclosed list")

    assert _find_global_config_file([tmp_path]) is None


def test_find_global_config_file(fmuconfig_with_output: Path) -> None:
    """Tests finding the global variables file in fmuconfig."""
    tmp_path = fmuconfig_with_output
    some_dir = tmp_path / "some_dir"
    some_dir.mkdir()
    some_file = some_dir / "some_file"
    some_file.touch()
    does_not_exist = tmp_path / "bad"
    assert _find_global_config_file([does_not_exist, some_dir, some_file]) is None

    fmuconfig_input = fmuconfig_with_output / "fmuconfig/input"
    global_variables = fmuconfig_input / "global_master_config.yml"
    shutil.copy(global_variables, fmuconfig_input / "global_master_config_pred.yml")

    assert isinstance(
        _find_global_config_file(
            [does_not_exist, some_dir, some_file, fmuconfig_input]
        ),
        GlobalConfiguration,
    )


def test_find_global_variables_file_not_there(tmp_path: Path) -> None:
    """Tests finding the global variables file if it is not present."""
    assert _find_global_variables_file([]) is None
    assert _find_global_variables_file([tmp_path]) is None
    assert _find_global_variables_file([tmp_path / "dne"]) is None


def test_find_global_variables_file_malformed_raises_validation_error(
    tmp_path: Path,
) -> None:
    """Tests that malformed global variables file raises ValidationError."""
    with open(tmp_path / "global_variables.yml", "w") as f:
        f.write("foo: bar")
    with pytest.raises(ValidationError):
        _find_global_variables_file([tmp_path])


def test_find_global_variables_file_returns_none_when_not_found(
    fmuconfig_with_output: Path,
) -> None:
    """Tests that None is returned when no global variables file exists."""
    tmp_path = fmuconfig_with_output
    some_dir = tmp_path / "some_dir"
    some_dir.mkdir()
    does_not_exist = tmp_path / "bad"
    assert _find_global_variables_file([does_not_exist, some_dir]) is None


def test_find_global_variables_file_skips_invalid_yaml_and_continues(
    tmp_path: Path,
) -> None:
    """Tests that function skips files with YAML parse errors and continues."""
    invalid_yaml_file = tmp_path / "global_variables.yml"
    with open(invalid_yaml_file, "w") as f:
        f.write("key: [unclosed list")

    assert _find_global_variables_file([tmp_path]) is None


def test_find_global_variables_file_raises_on_empty_file(
    fmuconfig_with_output: Path,
) -> None:
    """Tests that ValidationError is raised for empty/invalid file."""
    tmp_path = fmuconfig_with_output
    some_file = tmp_path / "some_file"
    some_file.touch()
    with pytest.raises(ValidationError):
        _find_global_variables_file([some_file])


def test_find_global_variables_file_returns_valid_config(
    fmuconfig_with_output: Path,
) -> None:
    """Tests finding a valid global variables file in fmuconfig."""
    assert isinstance(
        _find_global_variables_file([fmuconfig_with_output / "fmuconfig/output"]),
        GlobalConfiguration,
    )


def test_find_global_config_does_not_exist(tmp_path: Path) -> None:
    """Tests that find_global_config returns None if no config exists."""
    assert find_global_config(tmp_path) is None
    assert find_global_config(tmp_path, strict=False) is None


def test_find_global_config_from_input_non_strict(
    fmuconfig_with_input: Path,
) -> None:
    """Tests finding a global config from the input dir."""
    tmp_path = fmuconfig_with_input
    cfg = find_global_config(tmp_path, strict=False)
    assert isinstance(cfg, GlobalConfiguration)

    # Validate all fields present
    assert cfg.access.asset.name == "Drogon"
    assert cfg.access.classification == "internal"
    assert cfg.masterdata.smda.country[0].identifier == "Norway"
    assert cfg.masterdata.smda.discovery[0].short_identifier == "DROGON"
    assert cfg.masterdata.smda.field[0].identifier == "DROGON"
    assert cfg.masterdata.smda.coordinate_system.identifier == "ST_WGS84_UTM37N_P32637"
    assert (
        cfg.masterdata.smda.stratigraphic_column.identifier
        == "DROGON_HAS_NO_STRATCOLUMN"
    )
    assert cfg.model.name == "Drogon"
    assert cfg.stratigraphy is not None
    assert cfg.stratigraphy["TopVolantis"].name == "VOLANTIS GP. Top"


def test_find_global_config_from_input_strict(
    fmuconfig_with_input: Path,
) -> None:
    """Tests finding a global config with 'Drogon' in it raises."""
    tmp_path = fmuconfig_with_input
    with pytest.raises(
        InvalidGlobalConfigurationError, match="Invalid name in 'model': Drogon"
    ):
        find_global_config(tmp_path)


def test_find_global_config_extra_output_paths(
    fmuconfig_with_output: Path,
) -> None:
    """Tests finding global variables with extra output paths."""
    tmp_path = fmuconfig_with_output
    base_path = tmp_path / "some_dir"
    base_path.mkdir()

    cfg = find_global_config(
        base_path,
        extra_output_paths=[tmp_path / "fmuconfig/output/global_variables.yml"],
        strict=False,
    )
    assert isinstance(cfg, GlobalConfiguration)

    with pytest.raises(
        InvalidGlobalConfigurationError, match="Invalid name in 'model': Drogon"
    ):
        find_global_config(
            base_path,
            extra_output_paths=[tmp_path / "fmuconfig/output/global_variables.yml"],
        )
    assert isinstance(cfg, GlobalConfiguration)


def test_find_global_config_extra_input_paths(
    fmuconfig_with_input: Path,
) -> None:
    """Tests finding global variables with extra input paths."""
    tmp_path = fmuconfig_with_input
    base_path = tmp_path / "some_dir"
    base_path.mkdir()

    cfg = find_global_config(
        base_path,
        extra_input_dirs=[tmp_path / "fmuconfig/input"],
        strict=False,
    )
    assert isinstance(cfg, GlobalConfiguration)

    with pytest.raises(
        InvalidGlobalConfigurationError, match="Invalid name in 'model': Drogon"
    ):
        find_global_config(
            base_path,
            extra_input_dirs=[tmp_path / "fmuconfig/input"],
        )
    assert isinstance(cfg, GlobalConfiguration)



"""Tests for fmu.settings.resources.managers."""

import json
import shutil
from pathlib import Path
from typing import Self
from unittest.mock import patch

import pytest
from pydantic import BaseModel

from fmu.settings._fmu_dir import ProjectFMUDirectory
from fmu.settings._resources.lock_manager import LockManager
from fmu.settings._resources.pydantic_resource_manager import PydanticResourceManager


class A(BaseModel):
    """A test Pydantic class."""

    foo: str


class AManager(PydanticResourceManager[A]):
    """A test Pydantic resource manager."""

    def __init__(self: Self, fmu_dir: ProjectFMUDirectory) -> None:
        """Initializer."""
        super().__init__(fmu_dir, A)

    @property
    def relative_path(self: Self) -> Path:
        """Relative path."""
        return Path("foo.json")


def test_pydantic_resource_manager_implementation(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that derived classes must implement 'relative_path'."""

    class Manager(PydanticResourceManager[A]):
        """A test Pydantic resource manager."""

        def __init__(self: Self, fmu_dir: ProjectFMUDirectory) -> None:
            """Initializer."""
            super().__init__(fmu_dir, A)

    manager = Manager(fmu_dir)
    with pytest.raises(NotImplementedError):
        _ = manager.relative_path


def test_pydantic_resource_manager_init(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that initialization of a Pydantic resource manager is as expected."""
    a = AManager(fmu_dir)
    assert a.fmu_dir == fmu_dir
    assert a.model_class == A

    a_path = fmu_dir.path / "foo.json"
    assert a.path == a_path
    assert a.exists is False
    assert a._cache is None

    with pytest.raises(
        FileNotFoundError,
        match=f"Resource file for 'AManager' not found at: '{a_path}'",
    ):
        a.load()


def test_pydantic_resource_manager_save(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests saving a Pydantic resource that does not yet exist."""
    a = AManager(fmu_dir)
    a_model = A(foo="bar")

    a.save(a_model)

    assert a.exists
    assert a._cache == a_model
    with open(a.path, encoding="utf-8") as f:
        a_dict = json.loads(f.read())

    assert a_model == A.model_validate(a_dict)


def test_pydantic_resource_manager_save_raises_when_locked(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests saving raises when another process holds the lock."""
    lock = LockManager(fmu_dir)
    with (
        patch("socket.gethostname", return_value="other-host"),
        patch("os.getpid", return_value=12345),
    ):
        lock.acquire()

    a = AManager(fmu_dir)
    with pytest.raises(PermissionError, match="Cannot write to .fmu directory"):
        a.save(A(foo="bar"))

    with (
        patch("socket.gethostname", return_value="other-host"),
        patch("os.getpid", return_value=12345),
    ):
        lock.release()


def test_pydantic_resource_manager_load(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests loading a Pydantic resource."""
    a = AManager(fmu_dir)
    a_model = A(foo="bar")
    a.save(a_model)
    assert a.load() == a_model
    assert a._cache == a_model


def test_pydantic_resource_manager_load_force_true(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests loading a Pydantic resource with force=True."""
    a = AManager(fmu_dir)
    a_model = A(foo="bar")
    a.save(a_model)
    assert a.load() == a_model
    assert a._cache == a_model

    a_shadow = AManager(fmu_dir)
    a_shadow_model = A(foo="baz")
    a_shadow.save(a_shadow_model)

    assert a_shadow.load() == a_shadow_model
    assert a_shadow._cache == a_shadow_model

    assert a.load() == a_model
    assert a._cache == a_model

    assert a.load(force=True) == a_shadow_model
    assert a._cache == a_shadow_model


def test_pydantic_resource_manager_load_cache_false(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests loading a Pydantic resource with cache=False."""
    a = AManager(fmu_dir)
    a_model = A(foo="bar")
    a.save(a_model)

    a_shadow = AManager(fmu_dir)
    assert a_shadow.load(store_cache=False) == a_model
    assert a_shadow._cache is None

    assert a_shadow.load(store_cache=True) == a_model
    assert a_shadow._cache == a_model


def test_pydantic_resource_manager_load_force_true_cache_false(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests loading a Pydantic resource with force=True, cache=False."""
    a = AManager(fmu_dir)
    a_model = A(foo="bar")
    a.save(a_model)

    a_shadow = AManager(fmu_dir)
    a_shadow_model = A(foo="baz")
    a_shadow.save(a_shadow_model)

    assert a.load(force=True, store_cache=False) == a_shadow_model
    assert a._cache is a_model


def test_pydantic_resource_manager_loads_invalid_model(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Tests loading a Pydantic resource."""
    a = AManager(fmu_dir)
    a_model = A(foo="bar")
    a.save(a_model)

    a_dict = a_model.model_dump()
    a_dict["foo"] = 0

    fmu_dir.write_text_file(a.path, json.dumps(a_dict))

    with pytest.raises(
        ValueError, match=r"Invalid content in resource file[\s\S]*input_value=0"
    ):
        a.load(force=True)


def test_pydantic_resource_manager_save_does_not_cache_when_disabled(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Saving without cache enabled should not create cache artifacts."""
    original_default = AManager.automatic_caching
    AManager.automatic_caching = False
    cache_root = fmu_dir.path / "cache"
    try:
        if cache_root.exists():
            shutil.rmtree(cache_root)
        a = AManager(fmu_dir)
        a.save(A(foo="bar"))
    finally:
        AManager.automatic_caching = original_default

    assert not cache_root.exists()


def test_pydantic_resource_manager_save_stores_revision_when_enabled(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Saving with cache enabled should persist a revision snapshot."""
    a = AManager(fmu_dir)
    model = A(foo="bar")
    a.save(model)

    cache_root = fmu_dir.path / "cache"
    assert cache_root.is_dir()
    tag_path = cache_root / "CACHEDIR.TAG"
    assert tag_path.read_text(encoding="utf-8").startswith(
        "Signature: 8a477f597d28d172789f06886806bc55"
    )

    config_cache = cache_root / "foo"
    snapshots = list(config_cache.iterdir())
    assert len(snapshots) == 1
    snapshot = snapshots[0]
    assert snapshot.suffix == ".json"
    assert json.loads(snapshot.read_text(encoding="utf-8")) == model.model_dump()


def test_pydantic_resource_manager_revision_cache_trims_excess(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Revision caching should retain only the configured number of snapshots."""
    original_limit = fmu_dir.cache_max_revisions
    fmu_dir.cache_max_revisions = 5
    try:
        a = AManager(fmu_dir)
        a.save(A(foo="one"))
        a.save(A(foo="two"))
        a.save(A(foo="three"))
        a.save(A(foo="four"))
        a.save(A(foo="five"))
        a.save(A(foo="six"))
    finally:
        fmu_dir.cache_max_revisions = original_limit

    config_cache = fmu_dir.path / "cache" / "foo"
    snapshots = sorted(p.name for p in config_cache.iterdir())
    assert len(snapshots) == 5  # noqa: PLR2004

    contents = [
        json.loads((config_cache / name).read_text(encoding="utf-8"))["foo"]
        for name in snapshots
    ]
    assert contents == ["two", "three", "four", "five", "six"]



"""The model for config.json."""

import getpass
from datetime import UTC, datetime
from typing import Self

from pydantic import AwareDatetime, Field

from fmu.datamodels.fmu_results.fields import Access, Masterdata, Model
from fmu.settings import __version__
from fmu.settings.types import ResettableBaseModel, VersionStr  # noqa: TC001


class ProjectConfig(ResettableBaseModel):
    """The configuration file in a .fmu directory.

    Stored as config.json.
    """

    version: VersionStr
    created_at: AwareDatetime
    created_by: str
    masterdata: Masterdata | None = Field(default=None)
    model: Model | None = Field(default=None)
    access: Access | None = Field(default=None)
    cache_max_revisions: int = Field(default=5, ge=5)

    @classmethod
    def reset(cls: type[Self]) -> Self:
        """Resets the configuration to defaults.

        Returns:
            The new default Config object
        """
        return cls(
            version=__version__,
            created_at=datetime.now(UTC),
            created_by=getpass.getuser(),
            masterdata=None,
            model=None,
            access=None,
            cache_max_revisions=5,
        )



"""Contains models used for representing mappings within FMU."""

from pathlib import Path
from typing import Literal
from uuid import UUID

from pydantic import BaseModel, RootModel

from fmu.datamodels.enums import Content  # type: ignore

from ._enums import (
    DataEntrySource,
    MappingType,
    RelationType,
    TargetSystem,
)


class Source(BaseModel):
    name: str
    data_entry_source: DataEntrySource


class BaseMapping(BaseModel):
    """The base mapping containing the fields all mappings should contain.

    These fields will be contained in every individual mapping entry.
    """

    source_system: str
    target_system: TargetSystem
    mapping_type: MappingType


class IdentifierMapping(BaseMapping):
    """Base class for a one-to-one or many-to-one mapping of identifiers.

    This mapping represents takes some identifier from one source and correlates it to
    an identifier in a target. Most often this target will be some official masterdata
    store like SMDA.
    """

    source_id: str
    target_id: str
    target_uuid: UUID


class StratigraphyMapping(IdentifierMapping):
    """Represents a stratigraphy mapping.

    This is a mapping from stratigraphic aliases identifiers to an official
    identifier.
    """

    mapping_type: Literal[MappingType.stratigraphy] = MappingType.stratigraphy


class WellMapping(IdentifierMapping):
    """Represents a well mapping.

    This is a mapping from well aliases identifiers to an official
    identifier.
    """

    mapping_type: Literal[MappingType.well] = MappingType.well


class FaultMapping(IdentifierMapping):
    """Represents a fault mapping.

    This is a mapping from fault aliases identifiers to an official
    identifier.
    """

    mapping_type: Literal[MappingType.well]


class EntityReference(BaseModel):
    """Represents one entity we wish to related to naother entity.

    This is typically an object exported by dataio.
    """

    name: str
    uuid: UUID
    content: Content
    relative_path: Path
    absolute_path: Path


class RelationshipMapping(BaseMapping):
    """Base class for a mapping that represents a relationship between two entities."""

    source_entity: EntityReference
    target_entity: EntityReference
    relation_type: RelationType


class ParentChildMapping(BaseMapping):
    """A mapping between a child and their parent."""


class HierarchicalMapping(RelationshipMapping):
    """A mapping that contains a hierarchy."""


class Mappings(BaseModel):
    """A list of mappings under a mappings key in metadata or in a file on disk."""

    items: list[BaseMapping]


class OrderedMappings(Mappings):
    """Items in this list imply an ordering that is important in some context."""


MappingFile = RootModel[Mappings]
"""Represents a list of mappings contained in a text file."""



"""Utilities for storing revision snapshots of .fmu files."""

from __future__ import annotations

from datetime import UTC, datetime, timedelta
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, Final, Self
from uuid import uuid4

from fmu.settings._logging import null_logger

if TYPE_CHECKING:
    from fmu.settings._fmu_dir import FMUDirectoryBase

logger: Final = null_logger(__name__)

_CACHEDIR_TAG_CONTENT: Final = (
    "Signature: 8a477f597d28d172789f06886806bc55\n"
    "# This directory contains cached FMU files.\n"
    "# For information about cache directory tags, see:\n"
    "#	https://bford.info/cachedir/spec.html"
)


class CacheManager:
    """Stores complete file revisions under the `.fmu/cache` tree."""

    MIN_REVISIONS: ClassVar[int] = 5
    RETENTION_DAYS: ClassVar[int] = 30

    def __init__(
        self: Self,
        fmu_dir: FMUDirectoryBase,
        max_revisions: int = 5,
    ) -> None:
        """Initialize the cache manager.

        Args:
            fmu_dir: The FMUDirectory instance.
            max_revisions: Maximum number of revisions to retain. Default is 5.
                Values below 5 are set to 5.
        """
        self._fmu_dir = fmu_dir
        self._cache_root = Path("cache")
        self._max_revisions = max(self.MIN_REVISIONS, max_revisions)

    @property
    def max_revisions(self: Self) -> int:
        """Maximum number of revisions retained per resource."""
        return self._max_revisions

    @max_revisions.setter
    def max_revisions(self: Self, value: int) -> None:
        """Update the per-resource revision retention.

        Args:
            value: The new maximum number of revisions. Minimum value is 5.
                Values below 5 are set to 5.
        """
        self._max_revisions = max(self.MIN_REVISIONS, value)

    def store_revision(
        self: Self,
        resource_file_path: Path | str,
        content: str,
        encoding: str = "utf-8",
        skip_trim: bool = False,
    ) -> Path | None:
        """Write a full snapshot of the resource file to the cache directory.

        Args:
            resource_file_path: Relative path within the ``.fmu`` directory (e.g.,
                ``config.json``) of the resource file being cached.
            content: Serialized payload to store.
            encoding: Encoding used when persisting the snapshot. Defaults to UTF-8.
            skip_trim: If True, skip count-based trimming. Default is False.

        Returns:
            Absolute filesystem path to the stored snapshot.
        """
        resource_file_path = Path(resource_file_path)
        cache_dir = self._ensure_resource_cache_dir(resource_file_path)
        snapshot_name = self._snapshot_filename(resource_file_path)
        snapshot_path = cache_dir / snapshot_name

        cache_relative = self._cache_root / resource_file_path.stem
        self._fmu_dir.write_text_file(
            cache_relative / snapshot_name, content, encoding=encoding
        )
        logger.debug("Stored revision snapshot at %s", snapshot_path)

        if not skip_trim:
            self._trim(cache_dir)
        return snapshot_path

    def list_revisions(self: Self, resource_file_path: Path | str) -> list[Path]:
        """List existing snapshots for a resource file, sorted oldest to newest.

        Args:
            resource_file_path: Relative path within the ``.fmu`` directory (e.g.,
                ``config.json``) whose cache entries should be listed.

        Returns:
            A list of absolute `Path` objects sorted oldest to newest.
        """
        resource_file_path = Path(resource_file_path)
        cache_relative = self._cache_root / resource_file_path.stem
        if not self._fmu_dir.file_exists(cache_relative):
            return []
        cache_dir = self._fmu_dir.get_file_path(cache_relative)

        revisions = [p for p in cache_dir.iterdir() if p.is_file()]
        revisions.sort(key=lambda path: path.name)
        return revisions

    def _ensure_resource_cache_dir(self: Self, resource_file_path: Path) -> Path:
        """Create (if needed) and return the cache directory for resource file."""
        self._cache_root_path(create=True)
        resource_cache_dir_relative = self._cache_root / resource_file_path.stem
        return self._fmu_dir.ensure_directory(resource_cache_dir_relative)

    def _cache_root_path(self: Self, create: bool) -> Path:
        """Resolve the cache root, creating it and the cachedir tag if requested."""
        if create:
            cache_root = self._fmu_dir.ensure_directory(self._cache_root)
            self._ensure_cachedir_tag()
            return cache_root

        return self._fmu_dir.get_file_path(self._cache_root)

    def _ensure_cachedir_tag(self: Self) -> None:
        """Ensure the cache root complies with the Cachedir specification."""
        tag_path_relative = self._cache_root / "CACHEDIR.TAG"
        if self._fmu_dir.file_exists(tag_path_relative):
            return
        self._fmu_dir.write_text_file(tag_path_relative, _CACHEDIR_TAG_CONTENT)

    def _snapshot_filename(self: Self, resource_file_path: Path) -> str:
        """Generate a timestamped filename for the next snapshot."""
        timestamp = datetime.now(UTC).strftime("%Y%m%dT%H%M%S.%fZ")
        suffix = resource_file_path.suffix or ".txt"
        token = uuid4().hex[:8]
        return f"{timestamp}-{token}{suffix}"

    def _trim(self: Self, cache_dir: Path) -> None:
        """Remove the oldest snapshots until the retention limit is respected."""
        revisions = [p for p in cache_dir.iterdir() if p.is_file()]
        if len(revisions) <= self.max_revisions:
            return

        revisions.sort(key=lambda path: path.name)
        excess = len(revisions) - self.max_revisions
        for old_revision in revisions[:excess]:
            try:
                old_revision.unlink()
            except FileNotFoundError:
                continue

    def trim_by_age(
        self: Self, resource_file_path: Path | str, retention_days: int | None = None
    ) -> None:
        """Remove snapshots older than retention_days.

        Args:
            resource_file_path: Relative path within the ``.fmu`` directory (e.g.,
                ``logs/user_session_log.json``) whose cache entries should be trimmed.
            retention_days: Maximum age in days to retain snapshots.
                If None, uses CacheManager.RETENTION_DAYS (default: 30 days).
        """
        if retention_days is None:
            retention_days = self.RETENTION_DAYS
        revisions = self.list_revisions(resource_file_path)
        cutoff = datetime.now(UTC) - timedelta(days=retention_days)

        for revision in revisions:
            try:
                mtime_timestamp = revision.stat().st_mtime
                file_time = datetime.fromtimestamp(mtime_timestamp, tz=UTC)
            except (OSError, ValueError):
                logger.warning("Skipping file with unreadable mtime: %s", revision.name)
                continue

            if file_time < cutoff:
                revision.unlink()
                logger.debug("Deleted old revision: %s", revision)



"""Tests for the cache manager utilities."""

import os
from datetime import UTC, datetime, timedelta
from pathlib import Path

import pytest

from fmu.settings._fmu_dir import ProjectFMUDirectory
from fmu.settings._resources.cache_manager import (
    _CACHEDIR_TAG_CONTENT,
    CacheManager,
)


def _read_snapshot_names(config_cache: Path) -> list[str]:
    return sorted(p.name for p in config_cache.iterdir() if p.is_file())


def test_cache_manager_list_revisions_without_directory(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Listing revisions on a missing cache dir yields an empty list."""
    manager = CacheManager(fmu_dir)
    assert manager.list_revisions("foo.json") == []


def test_cache_manager_list_revisions_with_existing_snapshots(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Listing revisions returns sorted snapshot paths."""
    manager = CacheManager(fmu_dir)
    manager.store_revision("foo.json", "one")
    manager.store_revision("foo.json", "two")
    revisions = manager.list_revisions("foo.json")
    assert [path.name for path in revisions] == sorted(path.name for path in revisions)
    assert len(revisions) == 2  # noqa: PLR2004


def test_cache_manager_honours_existing_cachedir_tag(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Existing cachedir tags are preserved when storing revisions."""
    cache_root = fmu_dir.path / "cache"
    cache_root.mkdir(exist_ok=True)
    tag_path = cache_root / "CACHEDIR.TAG"
    tag_path.write_text("custom tag", encoding="utf-8")

    manager = CacheManager(fmu_dir)
    manager.store_revision("foo.json", '{"foo": "bar"}')

    assert tag_path.read_text(encoding="utf-8") == "custom tag"


def test_cache_manager_cache_root_helpers_create_tag(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Cache root helpers return consistent paths and create cachedir tags."""
    manager = CacheManager(fmu_dir)
    root = manager._cache_root_path(create=False)
    assert root == fmu_dir.get_file_path("cache")

    created = manager._cache_root_path(create=True)
    assert created == root

    tag_path = created / "CACHEDIR.TAG"
    assert tag_path.is_file()
    assert tag_path.read_text(encoding="utf-8") == _CACHEDIR_TAG_CONTENT


def test_cache_manager_uses_default_extension_for_suffixless_paths(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """Files without suffix get '.txt' snapshots."""
    manager = CacheManager(fmu_dir)
    snapshot = manager.store_revision("logs/entry", "payload")
    assert snapshot is not None
    assert snapshot.suffix == ".txt"
    assert snapshot.read_text(encoding="utf-8") == "payload"


def test_cache_manager_trim_handles_missing_files(
    fmu_dir: ProjectFMUDirectory,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Trimming gracefully handles concurrent removals."""
    manager = CacheManager(fmu_dir, max_revisions=CacheManager.MIN_REVISIONS)
    for i in range(CacheManager.MIN_REVISIONS + 2):
        manager.store_revision("foo.json", f"content_{i}")

    original_unlink = Path.unlink

    def flaky_unlink(self: Path, *, missing_ok: bool = False) -> None:
        if self.name.endswith(".json") and not getattr(flaky_unlink, "raised", False):
            flaky_unlink.raised = True  # type: ignore[attr-defined]
            original_unlink(self, missing_ok=missing_ok)
            raise FileNotFoundError
        original_unlink(self, missing_ok=missing_ok)

    monkeypatch.setattr(Path, "unlink", flaky_unlink)

    manager.store_revision("foo.json", "final")

    config_cache = fmu_dir.path / "cache" / "foo"
    assert len(_read_snapshot_names(config_cache)) == CacheManager.MIN_REVISIONS


def test_cache_manager_skip_trim_parameter(fmu_dir: ProjectFMUDirectory) -> None:
    """store_revision with skip_trim=True does not enforce count-based limit."""
    manager = CacheManager(fmu_dir, max_revisions=3)
    for i in range(5):
        manager.store_revision("foo.json", f"content_{i}", skip_trim=True)

    config_cache = fmu_dir.path / "cache" / "foo"
    snapshots = _read_snapshot_names(config_cache)
    assert len(snapshots) == 5  # noqa: PLR2004


def test_cache_manager_trim_by_age_removes_old_files(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """trim_by_age removes snapshots older than retention_days."""
    manager = CacheManager(fmu_dir)
    config_cache = fmu_dir.path / "cache" / "foo"
    config_cache.mkdir(parents=True, exist_ok=True)

    now = datetime.now(UTC)
    old_time = now - timedelta(days=35)
    recent_time = now - timedelta(days=10)

    old_filename = old_time.strftime("%Y%m%dT%H%M%S.%fZ") + "-abc12345.json"
    recent_filename = recent_time.strftime("%Y%m%dT%H%M%S.%fZ") + "-def67890.json"

    old_file = config_cache / old_filename
    recent_file = config_cache / recent_filename
    old_file.write_text("old content", encoding="utf-8")
    recent_file.write_text("recent content", encoding="utf-8")
    old_mtime = old_time.timestamp()
    recent_mtime = recent_time.timestamp()
    os.utime(old_file, (old_mtime, old_mtime))
    os.utime(recent_file, (recent_mtime, recent_mtime))

    manager.trim_by_age("foo.json", retention_days=30)

    remaining = _read_snapshot_names(config_cache)
    assert len(remaining) == 1
    assert remaining[0] == recent_filename


def test_cache_manager_trim_by_age_uses_default_retention(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """trim_by_age uses RETENTION_DAYS when retention_days is None."""
    manager = CacheManager(fmu_dir)
    config_cache = fmu_dir.path / "cache" / "foo"
    config_cache.mkdir(parents=True, exist_ok=True)

    now = datetime.now(UTC)
    old_time = now - timedelta(days=CacheManager.RETENTION_DAYS + 5)
    recent_time = now - timedelta(days=10)

    old_filename = old_time.strftime("%Y%m%dT%H%M%S.%fZ") + "-abc12345.json"
    recent_filename = recent_time.strftime("%Y%m%dT%H%M%S.%fZ") + "-def67890.json"

    old_file = config_cache / old_filename
    recent_file = config_cache / recent_filename
    old_file.write_text("old content", encoding="utf-8")
    recent_file.write_text("recent content", encoding="utf-8")
    old_mtime = old_time.timestamp()
    recent_mtime = recent_time.timestamp()
    os.utime(old_file, (old_mtime, old_mtime))
    os.utime(recent_file, (recent_mtime, recent_mtime))

    manager.trim_by_age("foo.json")

    remaining = _read_snapshot_names(config_cache)
    assert len(remaining) == 1
    assert remaining[0] == recent_filename


def test_cache_manager_trim_by_age_skips_malformed_files(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """trim_by_age skips files with unexpected format."""
    manager = CacheManager(fmu_dir)
    config_cache = fmu_dir.path / "cache" / "foo"
    config_cache.mkdir(parents=True, exist_ok=True)

    now = datetime.now(UTC)
    old_time = now - timedelta(days=35)
    old_filename = old_time.strftime("%Y%m%dT%H%M%S.%fZ") + "-abc12345.json"
    malformed_filename = "malformed_file.json"

    old_file = config_cache / old_filename
    malformed_file = config_cache / malformed_filename
    old_file.write_text("old content", encoding="utf-8")
    malformed_file.write_text("malformed", encoding="utf-8")
    old_mtime = old_time.timestamp()
    os.utime(old_file, (old_mtime, old_mtime))

    manager.trim_by_age("foo.json", retention_days=30)

    remaining = _read_snapshot_names(config_cache)
    assert len(remaining) == 1
    assert remaining[0] == malformed_filename


def test_cache_manager_trim_by_age_no_cache_directory(
    fmu_dir: ProjectFMUDirectory,
) -> None:
    """trim_by_age handles missing cache directory gracefully."""
    manager = CacheManager(fmu_dir)
    manager.trim_by_age("foo.json", retention_days=30)



"""Type annotations used in Pydantic models."""

from __future__ import annotations

from typing import Annotated, Self, TypeAlias

from pydantic import BaseModel, Field

VersionStr: TypeAlias = Annotated[
    str, Field(pattern=r"(\d+(\.\d+){0,2}|\d+\.\d+\.[a-z0-9]+\+[a-z0-9.]+)")
]


class ResettableBaseModel(BaseModel):
    """A Pydantic BaseModel that implements reset()."""

    @classmethod
    def reset(cls) -> Self:
        """Resets the model to an initial state."""
        raise NotImplementedError



"""Shared README content for .fmu directories."""

from textwrap import dedent
from typing import Final

PROJECT_README_CONTENT: Final[str] = dedent(
    """\
    This directory contains static configuration data for your FMU project.

    You should *not* manually modify files within this directory. Doing so may
    result in erroneous behavior or erroneous data in your FMU project.

    Changes to data stored within this directory must happen through the FMU
    Settings application.

    Run `fmu settings` to do this.
    """
)

USER_README_CONTENT: Final[str] = dedent(
    """\
    This directory contains static data and configuration elements used by some
    components in FMU. It may also contains sensitive access tokens that should not be
    shared with others.

    You should *not* manually modify files within this directory. Doing so may
    result in erroneous behavior by some FMU components.

    Changes to data stored within this directory must happen through the FMU
    Settings application.

    Run `fmu settings` to do this.
    """
)



"""Tests for ChangelogManager."""

import copy
import uuid
import warnings
from datetime import UTC, datetime, timedelta
from pathlib import Path
from typing import Any

import pandas as pd
import pytest
from fmu.datamodels.fmu_results.fields import StratigraphicColumn

from fmu.settings._fmu_dir import ProjectFMUDirectory
from fmu.settings._resources.changelog_manager import ChangelogManager
from fmu.settings.models._enums import ChangeType, FilterType
from fmu.settings.models.change_info import ChangeInfo
from fmu.settings.models.log import Filter, Log, LogFileName

DATE_TIME_NOW = datetime.now(UTC)


@pytest.fixture
def change_entry() -> ChangeInfo:
    """Returns a valid ChangeInfo object."""
    return ChangeInfo(
        timestamp=DATE_TIME_NOW,
        change_type=ChangeType.add,
        user="test",
        path=Path("/test_folder"),
        file="config.json",
        change="Added new field to smda masterdata.",
        hostname="hostname",
        key="masterdata",
    )


@pytest.fixture
def change_entry_list() -> list[ChangeInfo]:
    """Returns a list of valid ChangeInfo entries."""
    return [
        ChangeInfo(
            timestamp=DATE_TIME_NOW - timedelta(days=2),
            change_type=ChangeType.add,
            user="user_first_entry",
            path=Path("/path_first_entry"),
            file="config.json",
            change="Added new field.",
            hostname="hostname_first_entry",
            key="masterdata",
        ),
        ChangeInfo(
            timestamp=DATE_TIME_NOW - timedelta(days=1),
            change_type=ChangeType.update,
            user="user_second_entry",
            path=Path("/path_second_entry"),
            file="config.json",
            change="Updated field.",
            hostname="hostname_second_entry",
            key="masterdata",
        ),
        ChangeInfo(
            timestamp=DATE_TIME_NOW,
            change_type=ChangeType.remove,
            user="user_third_entry",
            path=Path("/path_third_entry"),
            file="config.json",
            change="Removed field.",
            hostname="hostname_third_entry",
            key="masterdata",
        ),
        ChangeInfo(
            timestamp=DATE_TIME_NOW,
            change_type=ChangeType.reset,
            user="user_fourth_entry",
            path=Path("/path_fourth_entry"),
            file="config.json",
            change="Reset field.",
            hostname="hostname_fourth_entry",
            key="masterdata",
        ),
    ]


def test_changelog_manager_instantiation(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests basic facts about the ChangelogManager."""
    changelog: ChangelogManager = ChangelogManager(fmu_dir)
    assert changelog.fmu_dir == fmu_dir
    assert changelog.relative_path == Path("logs") / LogFileName.changelog
    assert changelog.exists is False
    with pytest.raises(
        FileNotFoundError, match="Resource file for 'ChangelogManager' not found"
    ):
        changelog.load()

    assert changelog.model_class == Log[ChangeInfo]


def test_changelog_manager_add_entry(
    fmu_dir: ProjectFMUDirectory, change_entry: ChangeInfo
) -> None:
    """Tests adding a new entry to a non-existing changelog.

    This should create the changelog file and add the new entry at the end.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    assert changelog_resource.exists is False
    with pytest.raises(
        FileNotFoundError, match="Resource file for 'ChangelogManager' not found"
    ):
        changelog_resource.load()

    changelog_resource.add_log_entry(change_entry)

    assert changelog_resource.exists is True
    changelog = changelog_resource.load()
    assert changelog[0] == change_entry

    changelog_resource.add_log_entry(change_entry)
    changelog_resource.add_log_entry(change_entry)
    expected_log_entries = 3

    updated_changelog: Log[ChangeInfo] = changelog_resource.load()
    assert len(updated_changelog) == expected_log_entries


def test_changelog_manager_add_invalid_entry(
    fmu_dir: ProjectFMUDirectory, change_entry: ChangeInfo
) -> None:
    """Tests adding an invalid entry to an existing changelog.

    This should raise a Value error and not add the invalid log entry.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    changelog_resource.add_log_entry(change_entry)
    assert changelog_resource.exists

    change_entry_with_issues = copy.deepcopy(change_entry)
    del change_entry_with_issues.change_type

    with pytest.raises(
        ValueError, match=r"Invalid log entry added to 'Log\[ChangeInfo\]' "
    ):
        changelog_resource.add_log_entry(change_entry_with_issues)

    changelog: Log[ChangeInfo] = changelog_resource.load()
    assert len(changelog) == 1
    assert changelog[0] == change_entry


def test_changelog_manager_add_invalid_entry_no_file_created(
    fmu_dir: ProjectFMUDirectory, change_entry: ChangeInfo
) -> None:
    """Tests adding an invalid entry to a non-existing changelog.

    This should raise a Value error and not create the changelog file.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    assert changelog_resource.exists is False

    change_entry_with_issues = copy.deepcopy(change_entry)
    change_entry_with_issues.change_type = "invalid change_type"  # type: ignore

    with warnings.catch_warnings(record=True):
        warnings.simplefilter("always")
        with pytest.raises(
            ValueError, match=r"Invalid log entry added to 'Log\[ChangeInfo\]' "
        ):
            changelog_resource.add_log_entry(change_entry_with_issues)

    assert changelog_resource.exists is False


def test_changelog_filter_equal_operator(
    fmu_dir: ProjectFMUDirectory, change_entry_list: list[ChangeInfo]
) -> None:
    """Tests filtering changelog with == operator.

    The filter should return all changelog entries where the value of the field
    `field_name` equals the filter value.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    for change_entry in change_entry_list:
        changelog_resource.add_log_entry(change_entry)

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 4
    assert len(changelog) == expected_log_entries

    filter: Filter = Filter(
        field_name="change_type",
        filter_value=ChangeType.add,
        filter_type=FilterType.text,
        operator="==",
    )
    filtered_log = changelog_resource.filter_log(filter)
    assert len(filtered_log) == 1
    assert filtered_log[0] == changelog[0]

    filter = Filter(
        field_name="user",
        filter_value="user_second_entry",
        filter_type=FilterType.text,
        operator="==",
    )
    filtered_log = changelog_resource.filter_log(filter)
    assert len(filtered_log) == 1
    assert filtered_log[0] == changelog[1]

    filter = Filter(
        field_name="timestamp",
        filter_value=str(DATE_TIME_NOW),
        filter_type=FilterType.date,
        operator="==",
    )
    filtered_log = changelog_resource.filter_log(filter)
    expected_log_entries = 2
    assert len(filtered_log) == expected_log_entries
    assert all(entry.timestamp == DATE_TIME_NOW for entry in filtered_log)

    filter = Filter(
        field_name="change",
        filter_value="Changed field.",
        filter_type=FilterType.text,
        operator="==",
    )
    filtered_log = changelog_resource.filter_log(filter)
    assert len(filtered_log) == 0


def test_changelog_filter_not_equal_operator(
    fmu_dir: ProjectFMUDirectory, change_entry_list: list[ChangeInfo]
) -> None:
    """Tests filtering changelog with != operator.

    The filter should return all changelog entries where the value of the field
    `field_name` is not equal to the filter value.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    for change_entry in change_entry_list:
        changelog_resource.add_log_entry(change_entry)

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 4
    assert len(changelog) == expected_log_entries

    filter: Filter = Filter(
        field_name="change_type",
        filter_value=ChangeType.add,
        filter_type=FilterType.text,
        operator="!=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    expected_log_entries = 3
    assert len(filtered_log) == expected_log_entries
    assert all(entry.change_type is not ChangeType.add for entry in filtered_log)

    filter = Filter(
        field_name="user",
        filter_value="user_second_entry",
        filter_type=FilterType.text,
        operator="!=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    expected_log_entries = 3
    assert len(filtered_log) == expected_log_entries
    assert all(entry.user != "user_second_entry" for entry in filtered_log)

    filter = Filter(
        field_name="timestamp",
        filter_value=str(DATE_TIME_NOW),
        filter_type=FilterType.date,
        operator="!=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    expected_log_entries = 2
    assert len(filtered_log) == expected_log_entries
    assert all(entry.timestamp != DATE_TIME_NOW for entry in filtered_log)

    filter = Filter(
        field_name="key",
        filter_value="masterdata",
        filter_type=FilterType.text,
        operator="!=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    assert len(filtered_log) == 0


def test_changelog_filter_less_or_equal_operator(
    fmu_dir: ProjectFMUDirectory, change_entry_list: list[ChangeInfo]
) -> None:
    """Tests filtering changelog with <= operator.

    The filter should return all changelog entries where the value of the field
    `field_name` is less or equal to the filter value. Attempts to filter
    strings with the <= operator should raise an exception.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    for change_entry in change_entry_list:
        changelog_resource.add_log_entry(change_entry)

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 4
    assert len(changelog) == expected_log_entries

    filter: Filter = Filter(
        field_name="change_type",
        filter_value=ChangeType.add,
        filter_type=FilterType.text,
        operator="<=",
    )
    with pytest.raises(
        ValueError, match="Invalid filter operator <= applied to 'text' field"
    ):
        filtered_log = changelog_resource.filter_log(filter)

    filter = Filter(
        field_name="user",
        filter_value="user_third_entry",
        filter_type=FilterType.text,
        operator="<=",
    )
    with pytest.raises(
        ValueError, match="Invalid filter operator <= applied to 'text' field"
    ):
        filtered_log = changelog_resource.filter_log(filter)

    yesterday = DATE_TIME_NOW - timedelta(days=1)
    filter = Filter(
        field_name="timestamp",
        filter_value=str(yesterday),
        filter_type=FilterType.date,
        operator="<=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    expected_log_entries = 2
    assert len(filtered_log) == expected_log_entries
    assert all(entry.timestamp <= yesterday for entry in filtered_log)

    filter = Filter(
        field_name="timestamp",
        filter_value=str(DATE_TIME_NOW - timedelta(days=3)),
        filter_type=FilterType.date,
        operator="<=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    assert len(filtered_log) == 0


def test_changelog_filter_greater_or_equal_operator(
    fmu_dir: ProjectFMUDirectory, change_entry_list: list[ChangeInfo]
) -> None:
    """Tests filtering changelog with >= operator.

    The filter should return all changelog entries where the value of the field
    `field_name` is greater or equal to the filter value. Attempts to filter
    strings with the >= operator should raise an exception.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    for change_entry in change_entry_list:
        changelog_resource.add_log_entry(change_entry)

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 4
    assert len(changelog) == expected_log_entries

    filter: Filter = Filter(
        field_name="change_type",
        filter_value=ChangeType.add,
        filter_type=FilterType.text,
        operator=">=",
    )
    with pytest.raises(
        ValueError, match="Invalid filter operator >= applied to 'text' field"
    ):
        filtered_log = changelog_resource.filter_log(filter)

    filter = Filter(
        field_name="user",
        filter_value="user_third_entry",
        filter_type=FilterType.text,
        operator=">=",
    )
    with pytest.raises(
        ValueError, match="Invalid filter operator >= applied to 'text' field"
    ):
        filtered_log = changelog_resource.filter_log(filter)

    yesterday = DATE_TIME_NOW - timedelta(days=1)
    filter = Filter(
        field_name="timestamp",
        filter_value=str(yesterday),
        filter_type=FilterType.date,
        operator=">=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    expected_log_entries = 3
    assert len(filtered_log) == expected_log_entries
    assert all(entry.timestamp >= yesterday for entry in filtered_log)

    filter = Filter(
        field_name="timestamp",
        filter_value=str(DATE_TIME_NOW + timedelta(days=1)),
        filter_type=FilterType.date,
        operator=">=",
    )
    filtered_log = changelog_resource.filter_log(filter)
    assert len(filtered_log) == 0


def test_changelog_dataframe_cached_after_filtering(
    fmu_dir: ProjectFMUDirectory, change_entry: ChangeInfo
) -> None:
    """Tests that _cached_dataframe is set after filtering has been applied."""
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    assert changelog_resource._cached_dataframe is None
    changelog_resource.add_log_entry(change_entry)
    assert changelog_resource._cached_dataframe is None

    filter: Filter = Filter(
        field_name="change_type",
        filter_value=ChangeType.add,
        filter_type=FilterType.text,
        operator="==",
    )
    changelog_resource.filter_log(filter)
    assert changelog_resource._cached_dataframe is not None


def test_changelog_dataframe_cache_cleared(
    fmu_dir: ProjectFMUDirectory, change_entry: ChangeInfo
) -> None:
    """Tests that _cached_dataframe is cleared when the changelog is updated."""
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    assert changelog_resource._cached_dataframe is None

    changelog_resource._cached_dataframe = pd.DataFrame(["some_data"])
    assert changelog_resource._cached_dataframe is not None

    changelog_resource.add_log_entry(change_entry)
    assert changelog_resource._cached_dataframe is None


def test_log_update_to_changelog_when_flat_dict(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that updates of flat dictionaries are logged as expected.

    Checks that the key, change_type and change fields are logged with correct values.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)

    first_key = "first_key"
    first_value = "first_test_value"
    old_resource_dict = {first_key: first_value, "some_key": "some_value"}

    updated_value = "updated_value"
    added_key = "added_key"
    added_value = "added_value"
    updates: dict[str, Any] = {
        first_key: updated_value,
        added_key: added_value,
    }

    changelog_resource.log_update_to_changelog(
        updates, old_resource_dict, Path("config.json")
    )

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 2
    assert len(changelog) == expected_log_entries

    expected_change_string = (
        f"Updated field '{first_key}'. Old value: {first_value}"
        f" -> New value: {updated_value}"
    )
    assert changelog[0].change_type == ChangeType.update
    assert changelog[0].change == expected_change_string

    expected_change_string = f"Added field '{added_key}'. New value: {added_value}"
    assert changelog[1].change_type == ChangeType.add
    assert expected_change_string == changelog[1].change


def test_log_update_to_changelog_when_nested_dict(
    fmu_dir: ProjectFMUDirectory, masterdata_dict: dict[str, Any]
) -> None:
    """Tests that updates of nested dictionaries are logged as expected.

    Checks that the key, change_type and change fields are logged with correct values.
    """
    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)

    updated_country = [
        {
            "identifier": "Norge",
            "uuid": "00000000-0000-0000-0000-000000000000",
        }
    ]

    first_key = "smda.country"
    new_key = "new_key"
    new_nested_key = "new.nested.key"
    updates: dict[str, Any] = {
        first_key: updated_country,
        new_key: "new_value",
        new_nested_key: "new_nested_value",
    }

    changelog_resource.log_update_to_changelog(
        updates, masterdata_dict, Path("config.json")
    )

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 3
    assert len(changelog) == expected_log_entries

    expected_old_value = str(masterdata_dict["smda"]["country"])
    expected_change_string = (
        f"Updated field '{first_key}'. Old value: {expected_old_value}"
        f" -> New value: {str(updated_country)}"
    )

    assert changelog[0].change_type == ChangeType.update
    assert changelog[0].change == expected_change_string

    expected_change_string = f"Added field '{new_key}'. New value: new_value"
    assert changelog[1].change_type == ChangeType.add
    assert expected_change_string == changelog[1].change

    expected_change_string = (
        f"Added field '{new_nested_key}'. New value: new_nested_value"
    )
    assert changelog[2].change_type == ChangeType.add
    assert expected_change_string == changelog[2].change


def test_log_update_to_changelog_when_none_values(fmu_dir: ProjectFMUDirectory) -> None:
    """Tests that updates with None values are logged as expected.

    Checks that the key, change_type and change fields are logged with correct values.
    """
    first_key = "first_key"
    first_value = "first_test_value"
    second_key = "some_key"
    old_resource_dict = {first_key: first_value, second_key: None}

    updated_value = "updated_value"
    added_key = "added_key"
    updates: dict[str, Any] = {
        first_key: None,
        second_key: updated_value,
        added_key: None,
    }

    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    changelog_resource.log_update_to_changelog(
        updates, old_resource_dict, Path("config.json")
    )

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 3
    assert len(changelog) == expected_log_entries

    assert changelog[0].change_type == ChangeType.update
    expected_change_string = (
        f"Updated field '{first_key}'. Old value: {first_value}"
        f" -> New value: {str(None)}"
    )
    assert changelog[0].change == expected_change_string

    assert changelog[1].change_type == ChangeType.update
    expected_change_string = (
        f"Updated field '{second_key}'. Old value: {str(None)}"
        f" -> New value: {updated_value}"
    )
    assert changelog[1].change == expected_change_string

    assert changelog[2].change_type == ChangeType.add
    expected_change_string = f"Added field '{added_key}'. New value: {str(None)}"
    assert changelog[2].change_type == ChangeType.add
    assert expected_change_string == changelog[2].change


def test_log_update_to_changelog_when_base_model_values(
    fmu_dir: ProjectFMUDirectory, masterdata_dict: dict[str, Any]
) -> None:
    """Tests that updates to BaseModel objects are logged as expected.

    Checks that the key, change_type and change fields are logged with correct values.
    """
    strat_column = StratigraphicColumn(identifier="test_strat", uuid=uuid.uuid4())

    test_update = "smda.stratigraphic_column"
    test_add = "new.field"
    updates: dict[str, Any] = {test_update: strat_column, test_add: strat_column}

    changelog_resource: ChangelogManager = ChangelogManager(fmu_dir)
    changelog_resource.log_update_to_changelog(
        updates, masterdata_dict, Path("config.json")
    )

    changelog: Log[ChangeInfo] = changelog_resource.load()
    expected_log_entries = 2
    assert len(changelog) == expected_log_entries

    assert changelog[0].key == test_update
    assert changelog[0].change_type == ChangeType.update
    old_value = masterdata_dict["smda"]["stratigraphic_column"]
    expected_change_string = (
        f"Updated field '{test_update}'. Old value: {str(old_value)}"
        f" -> New value: {str(strat_column.model_dump())}"
    )
    assert expected_change_string == changelog[0].change

    assert changelog[1].key == test_add
    assert changelog[1].change_type == ChangeType.add
    expected_change_string = (
        f"Added field '{test_add}'. New value: {str(strat_column.model_dump())}"
    )
    assert expected_change_string == changelog[1].change
